# coding: utf8


from enum import Enum, unique


@unique
class IndexConvexityAdjustmentType(Enum):
    NONE = "None"
    BLACK_SCHOLES = "BlackScholes"
    REPLICATION = "Replication"
    LIBOR_SWAP_METHOD = "LiborSwapMethod"



from enum import Enum, unique
from typing import Union, List, Optional, Iterable

from ..._tools import make_convert_to_enum, EnumArgsParser, make_parse_enum


@unique
class AssetClass(Enum):
    """
    Asset class values to build 'filter' parameter in request for SymbolConversion content object.
    """

    COMMODITIES = "Commodities"
    EQUITY_OR_INDEX_OPTIONS = "EquityOrIndexOptions"
    BOND_AND_STIR_FUTURES_AND_OPTIONS = "BondAndSTIRFuturesAndOptions"
    WARRANTS = "Warrants"
    EQUITIES = "Equities"
    INDICES = "Indices"
    EQUITY_INDEX_FUTURES = "EquityIndexFutures"
    FUNDS = "Funds"
    CERTIFICATES = "Certificates"
    BONDS = "Bonds"
    RESERVE_CONVERTIBLE = "ReverseConvertible"
    MINI_FUTURE = "MiniFuture"
    FX_AND_MONEY = "FXAndMoney"


OptAssetClass = Optional[Union[str, List[str], AssetClass, List[AssetClass]]]

asset_class_enum_arg_parser = EnumArgsParser(
    parse=make_parse_enum(AssetClass), parse_to_enum=make_convert_to_enum(AssetClass)
)

search_all_category_by_asset_class = {
    AssetClass.COMMODITIES: "Commodities",
    AssetClass.EQUITY_OR_INDEX_OPTIONS: "Options",
    AssetClass.BOND_AND_STIR_FUTURES_AND_OPTIONS: "Exchange-Traded Rates",
    AssetClass.EQUITIES: "Equities",
    AssetClass.EQUITY_INDEX_FUTURES: "Futures",
    AssetClass.FUNDS: "Funds",
    AssetClass.BONDS: "Bond Pricing",
    AssetClass.FX_AND_MONEY: "FX & Money",
}

rcsasset_category_genealogy_by_asset_class = {
    AssetClass.WARRANTS: "A:AA",
    AssetClass.CERTIFICATES: "A:6N",
    AssetClass.INDICES: "I:17",
    AssetClass.RESERVE_CONVERTIBLE: "A:LE",
    AssetClass.MINI_FUTURE: "A:P6",
}


def _transform_to_string(values: Iterable, category: dict) -> str:
    return " ".join(f"'{category[value]}'" for value in values)


def create_asset_class_request_strings(asset_class: list) -> tuple:
    search_all_category_values = filter(
        lambda x: x in search_all_category_by_asset_class, asset_class
    )
    rcs_asset_category_values = filter(
        lambda x: x in rcsasset_category_genealogy_by_asset_class, asset_class
    )

    search_all_category_string_values = _transform_to_string(
        search_all_category_values, search_all_category_by_asset_class
    )

    search_all_rcs_asset_category_string_values = _transform_to_string(
        rcs_asset_category_values, rcsasset_category_genealogy_by_asset_class
    )

    search_all_category_string = ""
    rcs_asset_category_string = ""

    if search_all_category_string_values:
        search_all_category_string = (
            f"SearchAllCategoryv3 in ({search_all_category_string_values})"
        )

    if search_all_rcs_asset_category_string_values:
        rcs_asset_category_string = f"RCSAssetCategoryGenealogy in ({search_all_rcs_asset_category_string_values})"

    return search_all_category_string, rcs_asset_category_string



from refinitiv.data.content.ipa._curves._models._curve import Curve
from refinitiv.data.content.ipa._surfaces._models import Surface, SurfacePoint


def check_surface_curve(curve):
    assert isinstance(curve, Curve)
    assert curve.x.size > 0, f"Empty curve.x received"
    assert curve.y.size > 0, f"Empty curve.y received"


def check_surface_point(point):
    assert isinstance(point, SurfacePoint)
    assert point.x, "Empty point.x received"
    assert point.y, "Empty point.y received"
    assert point.z, "Empty point.z received"


def check_surface(surface):
    assert isinstance(surface, Surface)
    assert surface.x.size > 0, f"Empty surface.x received"
    assert surface.y.size > 0, f"Empty surface.y received"
    assert surface.z.size > 0, f"Empty surface.z received"



import allure
import pytest

from refinitiv.data.content.esg import universe
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_TITLES = ["Organization PermID", "Ric Code", "Common name"]
EXPECTED_COLUMN_NAMES = ["TR.OrganizationID", "TR.PrimaryRIC", "TR.CommonName"]


@allure.suite("Content object - ESG Universe")
@allure.feature("Content object - ESG Universe")
@allure.severity(allure.severity_level.CRITICAL)
class TestESGUniverse:
    @allure.title("Create ESG Universe definition object and get data - synchronously")
    @pytest.mark.caseid("C40663512")
    @pytest.mark.smoke
    def test_universe_get_data(self, open_platform_session):
        response = universe.Definition().get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, EXPECTED_TITLES)
        check_non_empty_response_data(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create ESG Universe definition object and get data with different use_field_names_in_headers params - asynchronously"
    )
    @pytest.mark.caseid("C40663507")
    async def test_universe_get_data_async(self, open_platform_session_async):
        first_response, second_response = await get_async_response_from_definitions(
            universe.Definition(use_field_names_in_headers=True),
            universe.Definition(use_field_names_in_headers=False),
        )
        check_response_status(
            response=first_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )
        check_non_empty_response_data(first_response)
        check_response_dataframe_contains_columns_names(
            first_response, EXPECTED_COLUMN_NAMES
        )

        check_response_status(
            response=second_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )
        check_non_empty_response_data(second_response)
        check_response_dataframe_contains_columns_names(
            second_response, EXPECTED_TITLES
        )

    @allure.title(
        "Create ESG Universe definition object and get data using closed session"
    )
    @pytest.mark.caseid("C40909207")
    def test_universe_get_data_using_closed_session(self, open_platform_session):
        session = open_platform_session
        session.close()
        with pytest.raises(ValueError) as error:
            universe.Definition().get_data()
        assert str(error.value) == "Session is not opened. Can't send any request"

    @allure.title(
        "Create ESG Universe definition object and get data without setting default session"
    )
    @pytest.mark.parametrize(
        "expected_error",
        ["No default session created yet. Please create a session first!"],
    )
    @pytest.mark.caseid("C40909209")
    def test_universe_get_data_without_setting_default_session(self, expected_error):
        with pytest.raises(AttributeError) as error:
            universe.Definition().get_data()
        assert str(error.value) == expected_error

    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-3784")
    @allure.title(
        "Create ESG Universe definition object and get data using invalid closure param"
    )
    @pytest.mark.parametrize(
        "closure, expected_error",
        [(0, "'closure' param must be 'string' object, not 'int'")],
    )
    @pytest.mark.caseid("C40909210")
    def test_universe_get_data_using_invalid_closure_param(
        self, open_platform_session, closure, expected_error
    ):
        with pytest.raises(TypeError) as error:
            universe.Definition(closure=closure).get_data()
        assert str(error.value) == expected_error



from ._enums import Package
from .._content_data_provider import ContentDataProvider
from .._error_parser import ErrorParser
from .._universe_content_validator import UniverseContentValidator
from ..._tools import (
    universe_arg_parser,
    make_enum_arg_parser,
    extend_params,
)
from ...delivery._data._data_provider import (
    RequestFactory,
    ValidatorContainer,
)


class EstimatesRequestFactory(RequestFactory):
    def get_query_parameters(self, *_, **kwargs) -> list:
        query_parameters = []
        universe = universe_arg_parser.get_str(kwargs.get("universe"), delim=",")
        query_parameters.append(("universe", universe))

        package = kwargs.get("package")
        if package is not None:
            package = package_estimates_arg_parser.get_str(package)
            query_parameters.append(("package", package))

        return query_parameters

    def extend_query_parameters(self, query_parameters, extended_params=None):
        return extend_params(query_parameters, extended_params)


package_estimates_arg_parser = make_enum_arg_parser(Package)

estimates_data_provider = ContentDataProvider(
    request=EstimatesRequestFactory(),
    validator=ValidatorContainer(content_validator=UniverseContentValidator()),
    parser=ErrorParser(),
)



from ...._curves._enums import CompoundingType, ExtrapolationMode
from ...._curves._bond_curves._enums import (
    BasisSplineSmoothModel,
    BusinessSector,
    CalendarAdjustment,
    CalibrationModel,
    CurveSubType,
    EconomicSector,
    Industry,
    IndustryGroup,
    InterestCalculationMethod,
    InterpolationMode,
    IssuerType,
    MainConstituentAssetClass,
    PriceSide,
    Rating,
    RatingScaleSource,
    ReferenceEntityType,
    Seniority,
)



import json

import allure
import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.errors import RDError
from tests.integration.content.ipa.contracts.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
    check_stream_state_and_df_from_stream,
    add_call_backs,
)
from tests.integration.content.ipa.contracts.repo.conftest import (
    repo_definition_01,
    repo_definition_02,
    on_response,
    repo_universe,
)
from tests.integration.helpers import get_async_response_from_definition


@allure.suite("Content object - Repo")
@allure.feature("Content object - Repo")
@allure.severity(allure.severity_level.CRITICAL)
class TestRepo:
    @allure.title(
        "Create two repo definition object with params in one session and get data"
    )
    @pytest.mark.caseid("36746892")
    def test_repo_get_data(self, open_session):
        events_list = []
        response_01 = repo_definition_01().get_data(
            on_response=lambda response, content_provider, session: on_response(
                response, session, events_list
            )
        )
        response_02 = repo_definition_02().get_data()

        assert "Response received for" in events_list, f"Events list is {events_list}"
        check_http_status_is_success_and_df_value_not_empty(response_01)
        check_http_status_is_success_and_df_value_not_empty(response_02)

    @allure.title("Create repo definition object without any parameters")
    @pytest.mark.caseid("36746910")
    def test_create_repo_without_any_parameters(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.repo.Definition().get_data()
        assert (
            str(error.value)
            == "Error code QPS-DPS.1024 | Invalid date. Provide either fields 'EndDate' or 'Tenor', not both or none."
        )

    @allure.title("Create repo definition asynchronously")
    @pytest.mark.caseid("36746933")
    async def test_repo_get_data_async(self, open_session_async):
        response = await get_async_response_from_definition(repo_definition_01())

        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create a repo stream with callbacks")
    @pytest.mark.caseid("36746936")
    def test_repo_get_stream(self, open_session):
        events_list = []
        stream = repo_definition_01().get_stream()

        add_call_backs(stream, events_list)

        stream.open()
        check_stream_state_and_df_from_stream(stream)
        assert "Response received for" in events_list, f"Events list is {events_list}"

    @allure.title("Create a repo definition with extended parameters")
    @pytest.mark.caseid("C39383379")
    def test_get_repo_with_extended_params(self, open_session):
        response = rdf.repo.Definition(
            underlying_instruments=[
                rdf.repo.UnderlyingContract(
                    instrument_type="Bond",
                    instrument_definition=rdf.bond.Definition(
                        instrument_code="US12345="
                    ),
                    pricing_parameters=rdf.repo.UnderlyingPricingParameters(
                        repo_parameters=rdf.repo.RepoParameters(
                            initial_margin_percent=20
                        )
                    ),
                )
            ],
            extended_params={
                "instrumentDefinition": {
                    "InstrumentTag": "my tag",
                    "startDate": "2020-08-17T00:00:00Z",
                    "endDate": "2020-08-22T00:00:00Z",
                    "underlyingInstruments": [
                        {
                            "instrumentType": "Bond",
                            "instrumentDefinition": {"instrumentCode": "US191450264="},
                            "pricingParameters": {
                                "repoParameters": {"initialMarginPercent": 50}
                            },
                        }
                    ],
                },
                "pricingParameters": {"marketDataDate": "2020-08-19T00:00:00Z"},
            },
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == repo_universe
        ), f"Extended params are applied incorrectly"
        assert response.data.df.InstrumentTag[0] == "my tag"



from refinitiv.data.content.ownership.investor import holdings


def test_holdings():
    try:
        definition = holdings.Definition(universe="")
    except Exception as e:
        assert False, str(e)



import allure
import pytest

from refinitiv.data.content.esg import standard_measures
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_COLUMN_NAMES = [
    "instrument",
    "periodenddate",
    "TR.TRESGCScore",
    "TR.TRESGScore",
    "TR.TRESGCControversiesScore",
    "TR.EnvironmentPillarScore",
    "TR.SocialPillarScore",
    "TR.GovernancePillarScore",
    "TR.TRESGResourceUseScore",
    "TR.TRESGEmissionsScore",
    "TR.TRESGInnovationScore",
    "TR.TRESGWorkforceScore",
    "TR.TRESGHumanRightsScore",
    "TR.TRESGCommunityScore",
    "TR.TRESGProductResponsibilityScore",
    "TR.TRESGManagementScore",
    "TR.TRESGShareholdersScore",
    "TR.TRESGCSRStrategyScore",
    "TR.CSRReportingScope",
    "TR.ESGPeriodLastUpdateDate",
]

EXPECTED_TITLES = [
    "Instrument",
    "Period End Date",
    "TRESG Combined Score",
    "TRESG Score",
    "ESG Controversies Score",
    "Environment Pillar Score",
    "Social Pillar Score",
    "Governance Pillar Score",
    "Resource Use Score",
    "Emissions Score",
    "CSR Sustainability Committee",
    "Integrated Strategy in MD&A",
    "Global Compact Signatory",
    "Stakeholder Engagement",
    "CSR Sustainability Reporting",
    "GRI Report Guidelines",
    "CSR Sustainability Report Global Activities",
    "CSR Sustainability External Audit",
    "CSR Sustainability External Auditor Name",
    "ESG Reporting Scope",
]


@allure.suite("Content object - ESG Standard Measures")
@allure.feature("Content object - ESG Standard Measures")
@allure.severity(allure.severity_level.CRITICAL)
class TestESGStandardMeasures:
    @allure.title("Create ESG Standard Measures definition object with valid params")
    @pytest.mark.parametrize(
        "universe,start,end,use_field_names_in_headers",
        [("4295904307", 0, -1, True), (["4295904307", "5000002406"], None, -2, False)],
    )
    @pytest.mark.caseid("C40466669")
    @pytest.mark.smoke
    def test_esg_standard_measures_definition_object_with_valid_params_and_get_data(
        self, universe, start, end, use_field_names_in_headers, open_platform_session
    ):
        response = standard_measures.Definition(
            universe=universe,
            start=start,
            end=end,
            use_field_names_in_headers=use_field_names_in_headers,
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_instrument_contains_expected_universes(
            response, universe
        )
        check_non_empty_response_data(response)
        check_dataframe_column_date_for_datetime_type(response)
        if use_field_names_in_headers is True:
            check_response_dataframe_contains_columns_names(
                response, EXPECTED_COLUMN_NAMES
            )
        else:
            check_response_dataframe_contains_columns_names(response, EXPECTED_TITLES)

    @allure.title(
        "Create ESG Standard Measures definition object with invalid params and get error"
    )
    @pytest.mark.caseid("C40466670")
    @pytest.mark.parametrize(
        "universe,expected_error",
        [
            (
                [],
                "Error code 400 | Validation error: Missing required parameter 'universe'",
            ),
            (
                ["INVALID"],
                "Error code 412 | Unable to resolve all requested identifiers. Requested items: ['INVALID']",
            ),
        ],
    )
    def test_esg_standard_measures_definition_object_with_invalid_params_and_get_error(
        self,
        universe,
        expected_error,
        open_platform_session,
    ):
        with pytest.raises(RDError) as error:
            standard_measures.Definition(universe=universe).get_data()
        assert str(error.value) == expected_error

    @allure.title(
        "Create ESG Standard Measures definition object with use mix valid and invalid data - asynchronously"
    )
    @pytest.mark.caseid("C40466671")
    @pytest.mark.parametrize(
        "universe,invalid_universe", [(["4295904307", "INVALID"], "INVALID")]
    )
    async def test_esg_standard_measures_definition_object_with_use_mix_valid_and_invalid_data(
        self, open_platform_session_async, universe, invalid_universe
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            standard_measures.Definition(universe=universe),
            standard_measures.Definition(universe=invalid_universe),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=None,
            expected_error_message="Failed to resolve identifiers ['INVALID']",
        )
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED_TWELVE,
            expected_error_message="Unable to resolve all requested identifiers. Requested items: ['INVALID']",
        )



# coding: utf8

from enum import Enum, unique


@unique
class CompoundingType(Enum):
    COMPOUNDED = "Compounded"
    CONTINUOUS = "Continuous"
    DISCOUNTED = "Discounted"
    MONEY_MARKET = "MoneyMarket"



from typing import Optional, TYPE_CHECKING

from ._stream_facade import Stream
from ...._tools import create_repr, validate_types

if TYPE_CHECKING:
    from ...._types import ExtendedParams
    from ...._core.session import Session


class Definition(object):
    """
    Class is designed to request streaming chains and decode it dynamically

    Parameters
    ----------
    name : str
        Single instrument name
    service : str, optional
        Name service
    skip_summary_links : bool, optional
        Store skip summary links
    skip_empty : bool, optional
        Store skip empty
    override_summary_links : int, optional
        Store the override number of summary links
    extended_params : dict, optional
        If necessary other parameters

    Examples
    --------
    >>> from refinitiv.data.content.pricing import chain
    >>> definition_chain = chain.Definition("0#.FTSE")
    """

    def __init__(
        self,
        name: str,
        service: Optional[str] = None,
        # option for chain constituents
        skip_summary_links: bool = True,
        skip_empty: bool = True,
        override_summary_links: Optional[int] = None,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(
            override_summary_links, [int, type(None)], "override_summary_links"
        )

        self._name = name
        self._service = service
        self._skip_summary_links = skip_summary_links
        self._skip_empty = skip_empty
        self._override_summary_links = override_summary_links
        self._extended_params = extended_params

    def __repr__(self):
        return create_repr(
            self,
            content=f"{{name='{self._name}'}}",
        )

    def get_stream(
        self,
        session: "Session" = None,
    ) -> Stream:
        """
        Return a chain.Stream object for the defined data

        Parameters
        ----------
        session : Session, optional
            The Session defines the source where you want to retrieve your data

        Returns
        -------
        chain.Stream

        Examples
        -------
        Create a chain.Stream object

        >>> from refinitiv.data.content.pricing import chain
        >>> definition_chain = chain.Definition("0#.FTSE")
        >>> chain_stream = definition_chain.get_stream()

        Open the Stream connection

        >>> from refinitiv.data.content.pricing import chain
        >>> definition_chain = chain.Definition("0#.FTSE")
        >>> chain_stream = definition_chain.get_stream()
        >>> chain_stream.open()

        Closes the Stream connection

        >>> from refinitiv.data.content.pricing import chain
        >>> definition_chain = chain.Definition("0#.FTSE")
        >>> chain_stream = definition_chain.get_stream()
        >>> chain_stream.open()
        >>> chain_stream.close()

        Call constituents

        >>> from refinitiv.data.content.pricing import chain
        >>> definition_chain = chain.Definition("0#.FTSE")
        >>> chain_stream = definition_chain.get_stream()
        >>> chain_stream.open()
        >>> chain_stream.constituents

        Call property is_chain

        >>> from refinitiv.data.content.pricing import chain
        >>> definition_chain = chain.Definition("0#.FTSE")
        >>> chain_stream = definition_chain.get_stream()
        >>> chain_stream.open()
        >>> chain_stream.is_chain
        """
        stream = Stream(
            name=self._name,
            session=session,
            service=self._service,
            skip_summary_links=self._skip_summary_links,
            skip_empty=self._skip_empty,
            override_summary_links=self._override_summary_links,
            extended_params=self._extended_params,
        )
        return stream



from types import SimpleNamespace


class Record(object):
    def __init__(self, name) -> None:
        super().__init__()
        self.name = name


stub_datetime = SimpleNamespace()
stub_datetime.strftime = lambda _: "datetime"



import datetime

from refinitiv.data.content import custom_instruments as ci

volume_based_udc = ci.manage.UDC(
    root="CC",
    months=ci.manage.Months(
        number_of_years=3,
        include_all_months=True,
        start_month=1,
    ),
    rollover=ci.manage.VolumeBasedRollover(
        method=ci.VolumeBasedRolloverMethod.VOLUME,
        number_of_days=1,
        join_at_day=1,
        roll_occurs_within_months=4,
        roll_on_expiry=True,
    ),
    spread_adjustment=ci.manage.SpreadAdjustment(
        adjustment="arithmetic",
        method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
        backwards=True,
    ),
)

ci_udc_response = {
    "id": "6b5d3003-622f-4a52-876e-ceb1ff2d773c",
    "symbol": "S)My_UDC_instrument.GEDTC-491962",
    "owner": "GEDTC-491962",
    "type": "udc",
    "udc": {
        "root": "CC",
        "months": {"numberOfYears": 3, "startMonth": 1, "includeAllMonths": True},
        "rollover": {
            "volumeBased": {
                "method": "volume",
                "numberOfDays": 1,
                "joinAtDay": 1,
                "rollOccursWithinMonths": 4,
                "rollOnExpiry": True,
            }
        },
        "spreadAdjustment": {
            "adjustment": "arithmetic",
            "method": "close-to-close",
            "backwards": True,
        },
    },
    "instrumentName": "9789",
    "holidays": [
        {"date": "1991-08-23", "reason": "Independence Day of Ukraine"},
        {"date": "2022-12-18", "reason": "Hanukkah"},
    ],
    "description": "fintech trading tool",
    "exchangeName": "5050",
    "currency": "GBP",
    "timeZone": "LON",
}

volume_based_dict_udc = ci_udc_response.get("udc")

day_based_udc = ci.manage.UDC(
    root="CL",
    months=ci.manage.Months(
        include_all_months=True,
        number_of_years=3,
    ),
    rollover=ci.manage.DayBasedRollover(
        method=ci.DayBasedRolloverMethod.DAYS_BEFORE_END_OF_MONTH,
        number_of_days=3,
        months_prior=1,
    ),
    spread_adjustment=ci.manage.SpreadAdjustment(
        adjustment="percentage",
        method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
        backwards=False,
    ),
)

day_based_dict_udc = {
    "root": "CL",
    "rollover": {
        "dayBased": {
            "method": "daysBeforeEndOfMonth",
            "numberOfDays": 3,
            "monthsPrior": 1,
        }
    },
    "spreadAdjustment": {
        "adjustment": "percentage",
        "method": "close-to-close",
        "backwards": False,
    },
    "months": {"numberOfYears": 3, "includeAllMonths": True},
}

manual_udc = ci.manage.UDC(
    root="CC",
    rollover=ci.manage.ManualRollover(
        ci.manage.ManualItem(month=7, year=2022, start_date="2022-02-01"),
        ci.manage.ManualItem(month=7, year=2021, start_date=datetime.date(2021, 3, 1)),
    ),
    spread_adjustment=ci.manage.SpreadAdjustment(
        adjustment="arithmetic",
        method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
        backwards=True,
    ),
)

manual_dict_udc = {
    "root": "CC",
    "rollover": {
        "manual": [
            {"month": 7, "year": 2022, "startDate": "2022-02-01T00:00:00.000000000Z"},
            {"month": 7, "year": 2021, "startDate": "2021-03-01T00:00:00.000000000Z"},
        ]
    },
    "spreadAdjustment": {
        "adjustment": "arithmetic",
        "method": "close-to-close",
        "backwards": True,
    },
}

basket_obj = ci.manage.Basket(
    constituents=[
        ci.manage.Constituent(ric="LSEG.L", weight=60),
        ci.manage.Constituent(ric="EPAM.N", weight=40),
    ],
    normalize_by_weight=True,
)

ci_basket_response = {
    "id": "c2d866ee-c8c7-4f0c-9391-789712e75436",
    "symbol": "S)My_Basket_instrument.GEDTC-491962",
    "owner": "GEDTC-491962",
    "type": "basket",
    "basket": {
        "constituents": [
            {"ric": "LSEG.L", "weight": 60.0},
            {"ric": "EPAM.N", "weight": 40.0},
        ],
        "normalizeByWeight": True,
    },
    "currency": "GBP",
    "instrumentName": "9789",
    "exchangeName": "5050",
    "holidays": [
        {"date": "1991-08-23", "reason": "Independence Day of Ukraine"},
        {"date": "2022-12-18", "reason": "Hanukkah"},
    ],
    "timeZone": "LON",
    "description": "fintech trading tool",
}

basket_dict = ci_basket_response.get("basket")

ci_create_formula_response = {
    "id": "6ded731a-24b4-4ab6-969f-9de4138b135b",
    "symbol": "S)My_formula_instrument.GEDTC-491962",
    "owner": "GEDTC-491962",
    "type": "formula",
    "formula": "GBP=*3",
    "currency": "GBP",
    "instrumentName": "9789",
    "exchangeName": "5050",
    "holidays": [
        {"date": "1991-08-23", "reason": "Independence Day of Ukraine"},
        {"date": "2022-12-18", "reason": "Hanukkah"},
    ],
    "timeZone": "LON",
    "description": "fintech trading tool",
}

ci_update_formula_response = {
    "id": "6ded731a-24b4-4ab6-969f-9de4138b135b",
    "symbol": "S)My_formula_instrument.GEDTC-491962",
    "owner": "GEDTC-491962",
    "type": "formula",
    "formula": "GBP=*3",
    "currency": "CAD",
    "instrumentName": "9789",
    "exchangeName": "5050",
    "holidays": [{"date": "2022-12-31", "reason": "New Year"}],
    "timeZone": "LON",
    "description": "fintech trading tool",
}



import typing as t

from httpx import Response, Request, HTTPTransport, AsyncHTTPTransport
from httpx import __version__ as _HTTPX_VERSION
from refinitiv.data._log import get_logger
from tenacity import (
    wait_exponential,
    Retrying,
    AsyncRetrying,
    stop_after_attempt,
    stop_after_delay,
)

from ...errors import RDError


class RequestRetryException(RDError):
    pass


class RequestTimeout(RequestRetryException):
    def __init__(
        self,
        timeout: float,
        total_time: float,
        attempts_count: int,
        attempts_max: int,
        url: str,
        method: str,
    ):
        self.timeout = timeout
        self.total_time = total_time
        self.attempts_count = attempts_count
        self.attempts_max = attempts_max
        self.url = url
        self.method = method
        super().__init__(
            -1,
            f"{method} request to {url} failed : "
            f"Timeout of {timeout} exceeded for retrying. "
            f"Total attempts: {attempts_count}",
        )


class RequestAttemptsExhausted(RequestRetryException):
    def __init__(
        self,
        timeout: float,
        total_time: float,
        attempts_count: int,
        attempts_max: int,
        url: str,
        method: str,
    ):
        self.timeout = timeout
        self.total_time = total_time
        self.attempts_count = attempts_count
        self.attempts_max = attempts_max
        self.url = url
        self.method = method
        super().__init__(
            -1,
            f"{method} request to {url} failed : "
            f"Max attempts count ({attempts_max}) exceeded for retrying. "
            f"Retry time: {total_time}",
        )


class _RetryTransportBase:
    def __init__(
        self,
        retrying_cls: t.Union[t.Type[Retrying], t.Type[AsyncRetrying]],
        total_timeout: float = 0,
        total_attempts: int = 10,
        on_statuses: t.Optional[t.Iterable[int]] = None,
        on_methods: t.Optional[t.Iterable[str]] = None,
        backoff_factor: float = 0,
    ):
        self._total_timeout = total_timeout
        self._total_attempts = total_attempts
        self._on_statuses: set = set(on_statuses or [])
        self._on_methods: set = set(map(str.upper, on_methods or []))
        self._backoff_factor = backoff_factor
        stop_condition = stop_after_attempt(total_attempts)
        self._logger = get_logger("RetryTransportBase")

        if self._total_timeout > 0:
            stop_condition |= stop_after_delay(self._total_timeout)
        self._retrying = retrying_cls(
            wait=wait_exponential(multiplier=backoff_factor),
            stop=stop_condition,
            reraise=True,
        )

    if "0.18" <= _HTTPX_VERSION < "0.20":

        def _handle_retry_exception(self, url, method, exc: RequestRetryException):
            stats = self._retrying.statistics
            if (
                self._total_timeout
                and 0 < self._total_timeout <= stats["delay_since_first_attempt"]
            ):
                raise RequestTimeout(
                    self._total_timeout,
                    stats["delay_since_first_attempt"],
                    stats["attempt_number"],
                    self._total_attempts,
                    url,
                    method,
                ) from exc
            elif stats["attempt_number"] >= self._total_attempts:
                raise RequestAttemptsExhausted(
                    self._total_timeout,
                    stats["delay_since_first_attempt"],
                    stats["attempt_number"],
                    self._total_attempts,
                    url,
                    method,
                ) from exc
            else:
                raise exc

        def _check_response(self, method: bytes, response: tuple):
            status_code, *_ = response
            if method.decode() in self._on_methods and status_code in self._on_statuses:
                raise RequestRetryException(
                    -1,
                    f"Request failed and will be retried: "
                    f"status code - {status_code}",
                )
            return response

    else:

        def _handle_retry_exception(self, request: Request, exc: RequestRetryException):
            stats = self._retrying.statistics
            if (
                self._total_timeout
                and 0 < self._total_timeout <= stats["delay_since_first_attempt"]
            ):
                raise RequestTimeout(
                    self._total_timeout,
                    stats["delay_since_first_attempt"],
                    stats["attempt_number"],
                    self._total_attempts,
                    request.url,
                    request.method,
                ) from exc
            elif stats["attempt_number"] >= self._total_attempts:
                raise RequestAttemptsExhausted(
                    self._total_timeout,
                    stats["delay_since_first_attempt"],
                    stats["attempt_number"],
                    self._total_attempts,
                    request.url,
                    request.method,
                ) from exc
            else:
                raise exc

        def _check_response(self, request: Request, response: Response):
            if (
                request.method in self._on_methods
                and response.status_code in self._on_statuses
            ):
                raise RequestRetryException(
                    -1,
                    f"Request failed and will be retried: "
                    f"status code - {response.status_code}",
                )
            return response


class RetryTransport(HTTPTransport, _RetryTransportBase):
    """
    Synchronous HTTP Transport with retry functionality

    Synchronous HTTP Transport with retry functionality that uses formula
     2^x * backoff_factor. Retry happens only if method is in the on_methods iterable
     and status code is in on_statuses iterable and/or if other network errors
     happen(connections issues, timeouts, etc).
     Otherwise the response is returned as is.

    Attributes
    ----------
    total_timeout : float
        Maximum time to keep retrying the request
    total_attempts: int
        Maximum count of attempts to try the request
    on_statuses: Iterable[int]
        Iterable of http status codes on which the retry should happen
    on_methods: Iterable[str]
        Iterable of http methods on which the retry should happen
    backoff_factor: float
        Multiplier to the exponential function
    **transport_kwargs: dict
        Keyword arguments to HTTPTransport
    Examples
    --------
        >>> from httpx import Client
        >>> transport = RetryTransport(total_timeout=10,
        >>>                            total_attempts=5,
        >>>                            on_statuses=[429, 501],
        >>>                            on_methods=['POST', 'GET'],
        >>>                            backoff_factor=2)
        >>> client = Client(transport=transport)

    """

    def __init__(
        self,
        total_timeout: float = 0,
        total_attempts: int = 10,
        on_statuses: t.Optional[t.Iterable[int]] = None,
        on_methods: t.Optional[t.Iterable[str]] = None,
        backoff_factor: float = 0,
        **transport_kwargs,
    ):
        HTTPTransport.__init__(self, **transport_kwargs)
        _RetryTransportBase.__init__(
            self,
            Retrying,
            total_timeout,
            total_attempts,
            on_statuses,
            on_methods,
            backoff_factor,
        )

    if "0.18" <= _HTTPX_VERSION < "0.20":

        def _handle_request(self, method: bytes, *args, **kwargs) -> tuple:
            return self._check_response(
                method, super().handle_request(method, *args, **kwargs)
            )

        def handle_request(
            self,
            method: bytes,
            url: t.Tuple[bytes, bytes, t.Optional[int], bytes],
            *args,
            **kwargs,
        ) -> tuple:
            try:
                return self._retrying(
                    self._handle_request, method, url, *args, **kwargs
                )
            except RequestRetryException as exc:
                self._handle_retry_exception(url, method, exc)

    else:  # For version 0.20+

        def _handle_request(self, request: Request) -> Response:
            self._logger.debug(f"Sending request to {request.url}")
            return self._check_response(
                request, super().handle_request(request=request)
            )

        def handle_request(self, request: Request) -> Response:
            try:
                return self._retrying(self._handle_request, request)
            except RequestRetryException as exc:
                self._handle_retry_exception(request, exc)


class RetryAsyncTransport(AsyncHTTPTransport, _RetryTransportBase):
    """
    Asynchronous HTTP Transport with retry functionality

    Asynchronous HTTP Transport with retry functionality that uses formula
     2^x * backoff_factor. Retry happens only if method is in the on_methods iterable
     and status code is in on_statuses iterable and/or if other network errors
     happen(connections issues, timeouts, etc).
     Otherwise the response is returned as is.

    Attributes
    ----------
    total_timeout : float
        Maximum time to keep retrying the request
    total_attempts: int
        Maximum count of attempts to try the request
    on_statuses: Iterable[int]
        Iterable of http status codes on which the retry should happen
    on_methods: Iterable[str]
        Iterable of http methods on which the retry should happen
    backoff_factor: float
        Multiplier to the exponential function
    **transport_kwargs: dict
        Keyword arguments to AsyncHTTPTransport
    Examples
    --------
        >>> from httpx import AsyncClient
        >>> transport = RetryAsyncTransport(total_timeout=10,
        >>>                                 total_attempts=5,
        >>>                                 on_statuses=[429, 501],
        >>>                                 on_methods=['POST', 'GET'],
        >>>                                 backoff_factor=2)
        >>> client = AsyncClient(transport=transport)

    """

    def __init__(
        self,
        total_timeout: float = 0,
        total_attempts: int = 10,
        on_statuses: t.Optional[t.Iterable[int]] = None,
        on_methods: t.Optional[t.Iterable[str]] = None,
        backoff_factor: float = 0,
        **transport_kwargs,
    ):
        AsyncHTTPTransport.__init__(self, **transport_kwargs)
        _RetryTransportBase.__init__(
            self,
            AsyncRetrying,
            total_timeout,
            total_attempts,
            on_statuses,
            on_methods,
            backoff_factor,
        )

    if "0.18" <= _HTTPX_VERSION < "0.20":

        async def _handle_async_request(self, method: bytes, *args, **kwargs) -> tuple:
            return self._check_response(
                method, await super().handle_async_request(method, *args, **kwargs)
            )

        async def handle_async_request(
            self,
            method: bytes,
            url: t.Tuple[bytes, bytes, t.Optional[int], bytes],
            *args,
            **kwargs,
        ) -> tuple:
            try:
                return await self._retrying(
                    self._handle_async_request, method, url, *args, **kwargs
                )
            except RequestRetryException as exc:
                self._handle_retry_exception(url, method, exc)

    else:

        async def _handle_async_request(self, request: Request) -> Response:
            return self._check_response(
                request, await super().handle_async_request(request=request)
            )

        async def handle_async_request(self, request: Request) -> Response:
            try:
                return await self._retrying(self._handle_async_request, request)
            except RequestRetryException as exc:
                self._handle_retry_exception(request, exc)



OHLC_COLUMNNS_NAMES = ["CLOSE", "COUNT", "HIGH", "LOW", "OPEN", "VOLUME"]


def check_dataframe_contains_ohlc_columns_and_rics(df, expected_rics):
    column_names = list(df.columns.values)
    rics = df.columns.name
    if isinstance(expected_rics, list):
        rics, column_names = df.axes[1].levels
    compare_arrays(OHLC_COLUMNNS_NAMES, column_names)
    compare_arrays(rics, expected_rics)


def compare_arrays(array1, array2):
    assert len(array1) == len(array2)
    for item in array1:
        assert item in array2



from unittest.mock import MagicMock

from refinitiv.data.session import desktop
from tests.unit.conftest import remove_dunder_methods


def test_desktop_session_scope():
    expected_attributes = [
        "Definition",
    ]
    testing_attributes = dir(desktop)
    testing_attributes = remove_dunder_methods(testing_attributes)
    assert expected_attributes == testing_attributes


def test_definition_keyword_arguments():
    app_key_mock = MagicMock(name="MOCK-app_key")

    definition = desktop.Definition(
        app_key=app_key_mock,
    )
    session = definition.get_session()
    assert session.app_key == app_key_mock


def test_definition_positional_arguments():
    session_name = "workspace"
    app_key_mock = MagicMock(name="MOCK-app_key")

    definition = desktop.Definition(
        session_name,
        app_key_mock,
    )
    session = definition.get_session()
    assert session.name == session_name
    assert session.app_key == app_key_mock


def test_definition_mix_of_positional_and_named_arguments():
    session_name = "workspace"
    app_key_mock = MagicMock(name="MOCK-app_key")

    definition = desktop.Definition(
        session_name,
        app_key_mock,
    )
    session = definition.get_session()
    assert session.name == session_name
    assert session.app_key == app_key_mock


def test_default_workspace_app_key_not_raise_error_if_exists(write_project_config):
    from refinitiv.data import _configure as configure

    configure.reload()

    config = {
        "sessions": {
            "default": "desktop.custom-session",
            "desktop": {
                "custom-session": {
                    "app-key": "DEFAULT_WORKSPACE_APP_KEY",
                }
            },
        },
    }
    write_project_config(config)
    configure.reload()
    try:
        definition = desktop.Definition()
        definition.get_session()
    except Exception as e:
        assert False, str(e)
    else:
        assert True



from ._session import Session
from ._session_cxn_type import SessionCxnType
from ._session_type import SessionType
from ..._tools import urljoin


class DesktopSession(Session):
    type = SessionType.DESKTOP

    def __init__(
        self,
        app_key,
        on_state=None,
        on_event=None,
        name="default",
        base_url=None,
        platform_path_rdp=None,
        platform_path_udf=None,
        handshake_url=None,
        token=None,
        dacs_position=None,
        dacs_application_id=None,
    ):
        super().__init__(
            app_key=app_key,
            on_state=on_state,
            on_event=on_event,
            token=token,
            dacs_position=dacs_position,
            dacs_application_id=dacs_application_id,
            name=name,
        )
        from os import getenv

        self._port = None
        self._udf_url = None
        self._timeout = self.http_request_timeout_secs

        # Detect DP PROXY url from CODEBOOK environment to manage multi user mode
        self._dp_proxy_base_url = getenv("DP_PROXY_BASE_URL")
        if self._dp_proxy_base_url:
            self._base_url = self._dp_proxy_base_url
        else:
            self._base_url = base_url

        self._platform_path_rdp = platform_path_rdp
        self._platform_path_udf = platform_path_udf
        self._handshake_url = handshake_url

        #   uuid is retrieved in CODEBOOK environment,
        #   it's used for DP-PROXY to manage multi-user mode
        self._uuid = getenv("REFINITIV_AAA_USER_ID")

        self._logger.debug(
            "".join(
                [
                    f"DesktopSession created with following parameters:",
                    f' app_key="{app_key}", name="{name}"',
                    f' base_url="{base_url}"' if base_url is not None else "",
                    f' platform_path_rdp="{platform_path_rdp}"'
                    if platform_path_rdp
                    else "",
                    f' platform_path_udf="{platform_path_udf}"'
                    if platform_path_udf
                    else "",
                    f' handshake_url="{handshake_url}"' if handshake_url else "",
                ]
            )
        )

    def _get_session_cxn_type(self) -> SessionCxnType:
        return SessionCxnType.DESKTOP

    def _get_udf_url(self):
        """
        Returns the url to request data to udf platform.
        """
        return urljoin(self._base_url, self._platform_path_udf)

    def _get_handshake_url(self):
        """
        Returns the url to handshake with the proxy.
        """
        return urljoin(self._base_url, self._handshake_url)

    def _get_base_url(self):
        return self._base_url

    def _get_rdp_url_root(self) -> str:
        if self._platform_path_rdp is None:
            raise ValueError(
                f"Can't find '{self.name}.platform-paths.rdp' "
                f"in config file. Please add this attribute."
            )
        url = urljoin(self._base_url, self._platform_path_rdp)
        return url

    def set_timeout(self, timeout):
        """
        Set the timeout for requests.
        """
        self._timeout = timeout

    def get_timeout(self):
        """
        Returns the timeout for requests.
        """
        return self._timeout

    def set_port_number(self, port_number):
        """
        Set the port number to reach Desktop API proxy.
        """
        self._port = port_number
        if port_number:
            try:
                protocol, path, default_port = self._base_url.split(":")
            except ValueError:
                protocol, path, *_ = self._base_url.split(":")
                default_port = ""

            try:
                url = ":".join([protocol, path, str(self._port)])
            except TypeError:
                url = ":".join([protocol, path, default_port])

            self._base_url = url
        else:
            self._udf_url = None

    def get_port_number(self):
        """
        Returns the port number
        """
        return self._port

    ############################################################
    #   multi-websockets support

    def get_omm_login_message(self):
        """return the login message for OMM 'key' section"""
        return {
            "Elements": {
                "AppKey": self.app_key,
                "ApplicationId": self._dacs_params.dacs_application_id,
                "Position": self._dacs_params.dacs_position,
                "Authorization": f"Bearer {self._access_token}",
            }
        }

    def get_rdp_login_message(self, stream_id):
        """return the login message for RDP"""
        return {
            "method": "Auth",
            "streamID": f"{stream_id:d}",
            "appKey": self.app_key,
            "authorization": f"Bearer {self._access_token}",
        }



# coding: utf8

from enum import Enum, unique


@unique
class AmortizationFrequency(Enum):
    EVERY12TH_COUPON = "Every12thCoupon"
    EVERY2ND_COUPON = "Every2ndCoupon"
    EVERY3RD_COUPON = "Every3rdCoupon"
    EVERY4TH_COUPON = "Every4thCoupon"
    EVERY_COUPON = "EveryCoupon"
    ONCE = "Once"



from typing import TYPE_CHECKING

from ...._object_definition import ObjectDefinition
from ._credit_curve_definition import CreditCurveDefinition
from ._credit_curve_parameters import CreditCurveParameters


if TYPE_CHECKING:
    from ......_types import OptStr
    from .._types import (
        CurveDefinition,
        CurveParameters,
    )


class RequestItem(ObjectDefinition):
    """
    Generates the credit curves for the definition provided

    Parameters
    ----------
    curve_definition : CreditCurveDefinition, optional

    curve_parameters : CreditCurveParameters, optional

    curve_tag : str, optional
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
    """

    def __init__(
        self,
        curve_definition: "CurveDefinition" = None,
        curve_parameters: "CurveParameters" = None,
        curve_tag: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.curve_definition = curve_definition
        self.curve_parameters = curve_parameters
        self.curve_tag = curve_tag

    @property
    def curve_definition(self):
        """
        :return: object CreditCurveDefinition
        """
        return self._get_object_parameter(CreditCurveDefinition, "curveDefinition")

    @curve_definition.setter
    def curve_definition(self, value):
        self._set_object_parameter(CreditCurveDefinition, "curveDefinition", value)

    @property
    def curve_parameters(self):
        """
        :return: object CreditCurveParameters
        """
        return self._get_object_parameter(CreditCurveParameters, "curveParameters")

    @curve_parameters.setter
    def curve_parameters(self, value):
        self._set_object_parameter(CreditCurveParameters, "curveParameters", value)

    @property
    def curve_tag(self):
        """
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
        :return: str
        """
        return self._get_parameter("curveTag")

    @curve_tag.setter
    def curve_tag(self, value):
        self._set_parameter("curveTag", value)



from .._enums import MainConstituentAssetClass, RiskType
from ...._object_definition import ObjectDefinition


class BaseDefinitionMixin(ObjectDefinition):
    @property
    def main_constituent_asset_class(self):
        return self._get_enum_parameter(
            MainConstituentAssetClass, "mainConstituentAssetClass"
        )

    @main_constituent_asset_class.setter
    def main_constituent_asset_class(self, value):
        self._set_enum_parameter(
            MainConstituentAssetClass, "mainConstituentAssetClass", value
        )

    @property
    def risk_type(self):
        return self._get_enum_parameter(RiskType, "riskType")

    @risk_type.setter
    def risk_type(self, value):
        self._set_enum_parameter(RiskType, "riskType", value)

    @property
    def base_currency(self):
        return self._get_parameter("baseCurrency")

    @base_currency.setter
    def base_currency(self, value):
        self._set_parameter("baseCurrency", value)

    @property
    def base_index_name(self):
        return self._get_parameter("baseIndexName")

    @base_index_name.setter
    def base_index_name(self, value):
        self._set_parameter("baseIndexName", value)

    @property
    def definition_expiry_date(self):
        return self._get_parameter("definitionExpiryDate")

    @definition_expiry_date.setter
    def definition_expiry_date(self, value):
        self._set_parameter("definitionExpiryDate", value)

    @property
    def is_fallback_for_fx_curve_definition(self):
        return self._get_parameter("isFallbackForFxCurveDefinition")

    @is_fallback_for_fx_curve_definition.setter
    def is_fallback_for_fx_curve_definition(self, value):
        self._set_parameter("isFallbackForFxCurveDefinition", value)

    @property
    def is_non_deliverable(self):
        return self._get_parameter("isNonDeliverable")

    @is_non_deliverable.setter
    def is_non_deliverable(self, value):
        self._set_parameter("isNonDeliverable", value)

    @property
    def name(self):
        return self._get_parameter("name")

    @name.setter
    def name(self, value):
        self._set_parameter("name", value)

    @property
    def quoted_currency(self):
        return self._get_parameter("quotedCurrency")

    @quoted_currency.setter
    def quoted_currency(self, value):
        self._set_parameter("quotedCurrency", value)

    @property
    def quoted_index_name(self):
        return self._get_parameter("quotedIndexName")

    @quoted_index_name.setter
    def quoted_index_name(self, value):
        self._set_parameter("quotedIndexName", value)

    @property
    def source(self):
        return self._get_parameter("source")

    @source.setter
    def source(self, value):
        self._set_parameter("source", value)


class BaseWithIdDefinitionMixin(BaseDefinitionMixin):
    @property
    def id(self):
        return self._get_parameter("id")

    @id.setter
    def id(self, value):
        self._set_parameter("id", value)



from ..._enums import (
    AdjustInterestToPaymentDate,
    AmortizationFrequency,
    AmortizationType,
    BusinessDayConvention,
    DateRollingConvention,
    DayCountBasis,
    Direction,
    Frequency,
    IndexAverageMethod,
    IndexCompoundingMethod,
    IndexConvexityAdjustmentIntegrationMethod,
    IndexConvexityAdjustmentMethod,
    IndexObservationMethod,
    IndexResetType,
    IndexSpreadCompoundingMethod,
    InterestCalculationConvention,
    InterestType,
    NotionalExchange,
    PriceSide,
    StubRule,
    TenorReferenceDate,
    SwaptionType,
    PremiumSettlementType,
)



import pytest
from mock import Mock

from refinitiv.data._core.session._session_provider import make_session_provider


def test_make_session_provider_returns_provider(session_type):
    # when
    provider = make_session_provider(session_type)

    # then
    assert provider is not None


def test_make_session_provider_raise_error_if_invalid_session_type():
    # then
    with pytest.raises(ValueError):
        # when
        make_session_provider("invalid-session-type")


def test_session_provider_create_class_by_params(session_type):
    # given
    grant = Mock()
    config = {"app-key": "test_app_key"}
    # when
    create_session = make_session_provider(session_type, config, grant)
    session = create_session()

    # then
    assert session is not None


def test_session_provider_create_class_by_params_and_kwargs(session_type):
    # given
    grant = Mock()
    config = {"app-key": "test_app_key"}
    # when
    create_session = make_session_provider(session_type, config, grant)
    session = create_session()

    # then
    assert session is not None



import pytest

from refinitiv.data._tools.templates import InvalidPlaceholderError
from refinitiv.data.discovery._search_templates.base import (
    extract_used_templates_from_placeholders_and_namespace,
    TargetTemplate,
    Target,
)
from refinitiv.data.discovery._search_templates.namespaces import Namespace
from tests.unit.content.search.search_templates.catshelterdb import CatsDatabaseTarget


class TestExtractUsedTemplatesFromPlaceholdersAndNamespace:
    def test_basic(self):
        ns = Namespace(
            _=Namespace(
                One=TargetTemplate(name="One"),
            ),
            user=Namespace(
                sub=Namespace(
                    Two=TargetTemplate(name="Two"),
                ),
                Three=TargetTemplate(name="Three"),
            ),
        )
        names = {"_.One.col", "user.sub.Two.val"}
        assert extract_used_templates_from_placeholders_and_namespace(names, ns) == {
            "_.One",
            "user.sub.Two",
        }

    def test_simplest(self):
        assert (
            extract_used_templates_from_placeholders_and_namespace(
                {"pre.A.val"},
                Namespace(_=Namespace(A=TargetTemplate(name="A"))),
            )
            == set()
        )

    def test_uknown_subtemplate(self):
        with pytest.raises(ValueError):
            extract_used_templates_from_placeholders_and_namespace(
                {"_.A", "_.B"},
                Namespace(_=Namespace(A=TargetTemplate(name="A"))),
            )


class CatSearchTemplate(TargetTemplate):
    _target_class = CatsDatabaseTarget


class TestTargetTemplate:
    def test_basic(self):
        birman_cats = CatSearchTemplate(
            name="Search Birman cats",
            view="cats",
            filter="breed == 'Birman'",
        )
        assert set(birman_cats.search().breed) == {"Birman"}

    def test_repr(self):
        assert (
            repr(TargetTemplate(name="TT")) == "<TargetTemplate for Target name='TT'>"
        )

    def test_placeholders(self):
        search_by_breed = CatSearchTemplate(
            name="Search Birman cats",
            view="cats",
            filter="breed == '#{breed}'",
        )
        assert set(search_by_breed.search(breed="Sphynx").breed) == {"Sphynx"}

    def test_subtemplate_namespace(self):
        body_type_cats_tpl = CatSearchTemplate(
            name="Cats by body type",
            ns=Namespace(
                locals=Namespace(
                    BreedsByBodyType=CatSearchTemplate(
                        name="Breeds by body_type",
                        view="breeds",
                        filter="body_type == '#{body_type}'",
                    ),
                ),
            ),
            view="cats",
            filter="breed in (#{locals.BreedsByBodyType.name|in})",
        )
        assert body_type_cats_tpl._subtemplates_used == {"locals.BreedsByBodyType"}
        assert (
            body_type_cats_tpl._search_kwargs(body_type="dwarf")["filter"]
            == "breed in ('Munchkin')"
        )
        assert set(body_type_cats_tpl.search(body_type="dwarf").name) == {
            "Elsa",
            "Maverick",
        }

        body_type_in_city = CatSearchTemplate(
            name="Cats by body type",
            ns=Namespace(
                user=Namespace(
                    SheltersInCity=CatSearchTemplate(
                        name="Shelters in city",
                        view="shelters",
                        filter="city == '#{city}'",
                    )
                ),
                locals=Namespace(
                    BreedsByBodyType=CatSearchTemplate(
                        name="Breeds by body_type",
                        view="breeds",
                        filter="body_type == '#{body_type}'",
                    ),
                ),
            ),
            optional_placeholders=["city"],
            view="cats",
            filter="breed in (#{locals.BreedsByBodyType.name|in}){{if city}} and shelter in (#{user.SheltersInCity.name|in}){{endif}}",
        )

        # TODO: this must work, but it does not. Need to develop more complex logic for used templates detection
        # assert body_type_in_city._search_kwargs(body_type="dwarf")["filter"] == \
        #     "breed in ('Munchkin')"
        # assert set(body_type_in_city.search(body_type="dwarf").name) == {
        #     "Elsa", "Maverick",
        # }

        assert (
            body_type_in_city._search_kwargs(body_type="dwarf", city="Paris")["filter"]
            == "breed in ('Munchkin') and shelter in ('01', '03')"
        )
        assert set(body_type_in_city.search(body_type="dwarf", city="Paris").name) == {
            "Maverick",
        }


class DummyTarget(Target):
    """DummyTarget, call does nothing"""

    def __init__(self):
        super().__init__()
        self.args_names = {"filter", "boost", "top", "skip"}


class DummyTemplate(TargetTemplate):
    _target_class = DummyTarget


class TestTargetTemplateKwargs:
    """Test TargetTemplate API without calling Target"""

    def test_raise_error_for_default_not_in_definition(self):
        with pytest.raises(Exception):
            TargetTemplate(bad_argument=None)

    @pytest.mark.skip("other exception in jinja")
    def test_invalid_placeholder_syntax(self):
        with pytest.raises(InvalidPlaceholderError):
            TargetTemplate(boost="ExchangeName xeq '#???'")

    def test_applies_substitution_variables(self):
        template = DummyTemplate(
            filter="AssetState ne '#{state}' and SearchAllCategoryv2 eq '#{cat}'",
            boost="ExchangeName xeq '#{name}'",
        )
        assert template._search_kwargs(state="state", cat="cat", name="name") == {
            "filter": "AssetState ne 'state' and SearchAllCategoryv2 eq 'cat'",
            "boost": "ExchangeName xeq 'name'",
        }

    @pytest.mark.skip("jinja does not support alternative syntax")
    def test_escape_symbol(self):
        template = DummyTemplate(boost="#name and ##name")
        assert template._search_kwargs(name="name") == {"boost": "name and #name"}

    @pytest.mark.skip("jinja does not support alternative syntax")
    def test_applies_non_braced_variables(self):
        template = DummyTemplate(boost="ExchangeName xeq '#name'")
        assert template._search_kwargs(name="name") == {
            "boost": "ExchangeName xeq 'name'",
        }

    def test_exception_raised_if_sub_value_is_not_there(self):
        template = DummyTemplate(boost="ExchangeName xeq '#{name}'")
        with pytest.raises(Exception):
            template._search_kwargs()

    def test_cant_redefine_parameter_with_placeholders(self):
        template = DummyTemplate(boost="ExchangeName xeq '#{name}'")
        with pytest.raises(Exception):
            template._search_kwargs(boost=None, name="")  # no exception

    def test_non_template_arguments(self):
        template = DummyTemplate(top=3, skip=2)
        assert template._search_kwargs() == {"top": 3, "skip": 2}

    def test_tunnels_other_arguments(self):
        template = DummyTemplate(top=3)
        assert template._search_kwargs(skip=2) == {"top": 3, "skip": 2}

    def test_exception_on_bad_kwarg(self):
        template = DummyTemplate()
        with pytest.raises(Exception):
            template._search_kwargs(bad_argument_name=None)

    def test_templates_variables_with_same_name_as_search_argument(self):
        with pytest.raises(Exception):
            DummyTemplate(boost="ExchangeName xeq '#{top}'")

    def test_default_values(self):
        template = DummyTemplate(
            boost="ExchangeName xeq '#{exchange}'",
            placeholders_defaults={"exchange": 5},
        )
        assert template._search_kwargs() == {
            "boost": "ExchangeName xeq '5'",
        }

    def test_search_kwargs_result_changes_must_not_change_template(self):
        template = DummyTemplate(top=3, skip=2)
        result = template._search_kwargs()
        result["top"] = 5
        assert template._search_kwargs()["top"] == 3

    def test_bad_pass_through(self):
        with pytest.raises(Exception):
            template = DummyTemplate(pass_through_defaults={"bad_pass_through": 5})



import sys
from concurrent.futures import Future
import pytest

from refinitiv.data.delivery import omm_stream
from refinitiv.data.delivery._stream import _OMMStream, StreamState
from tests.unit.conftest import StubSession


def test_omm_stream_definition_fields_is_list_of_strings():
    # given
    fields = "BID"

    # when
    definition = omm_stream.Definition(name="GBP=", fields=fields)

    # then
    assert isinstance(definition._fields, list)


def test_opened_future_set_true_after_get_status_event():
    # given
    session = StubSession()
    stream = _OMMStream(1, session, "BB.TO", None, "MarketByPrice", "ELEKTRON_DD")

    stream._state = StreamState.Opening

    # when
    stream._on_stream_status({}, {})

    # then
    assert stream._opened.is_set()


def test_opened_future_does_not_set_true_after_get_refresh_event():
    # given
    session = StubSession()
    stream = _OMMStream(1, session, "BB.TO", None, "MarketByPrice", "ELEKTRON_DD")

    stream._state = StreamState.Opening

    # when
    stream._on_stream_refresh({}, {})

    # then
    assert not stream._opened.is_set()


def test_opened_future_does_not_raise_invalid_state_error():
    # given
    session = StubSession(is_open=True)
    stream = _OMMStream(1, session, "BB.TO", None, "MarketByPrice", "ELEKTRON_DD")

    # when
    # open()
    stream._state = StreamState.Opening
    stream._on_stream_status({}, {})
    stream._state = StreamState.Opened

    try:
        stream._on_stream_complete({}, {})
    except Exception as e:
        assert False, str(e)
    else:
        # then
        assert True


def test_opened_future_does_not_raise_invalid_state_error_if_stream_got_close_message():
    # given
    session = StubSession(is_open=True)
    stream = _OMMStream(1, session, "BB.TO", None, "MarketByPrice", "ELEKTRON_DD")

    # when
    # open()
    stream._state = StreamState.Opening
    close_message = {
        "ID": 6,
        "Type": "Status",
        "Key": {"Service": "IDN_RDFNTS_CF", "Name": "INVALID"},
        "State": {
            "Stream": "Closed",
            "Data": "Suspect",
            "Code": "NotFound",
            "Text": "The record could not be found",
        },
    }
    stream._on_stream_status({}, close_message)
    stream._state = StreamState.Opened

    try:
        stream._on_stream_complete({}, {})
    except Exception as e:
        assert False, str(e)
    else:
        # then
        assert True



from enum import Enum, unique

import pytest

from refinitiv.data.content.news._news_data_provider import sort_order_news_arg_parser
from refinitiv.data.content.news._sort_order import SortOrder


@unique
class SomeEnum(Enum):
    Foo = "foo"
    Buzz = "buzz"


@pytest.mark.unit
def test_throw_error_when_convert_some_enum():
    some_enum = SomeEnum.Foo

    with pytest.raises(AttributeError):
        sort_order_news_arg_parser.get_str(some_enum)


@pytest.mark.unit
def test_throw_error_when_convert_incorrect_symbol_str():
    some_str = "foo"

    with pytest.raises(AttributeError):
        sort_order_news_arg_parser.get_str(some_str)


@pytest.mark.unit
def test_convert_symbol_type_to_correct_str():
    for sym_type in SortOrder:
        sym_str = sort_order_news_arg_parser.get_str(sym_type)

        assert sym_str == sym_type.value


@pytest.mark.unit
def test_convert_correct_symbol_str_to_correct_str():
    for sym_str, item in SortOrder.__members__.items():
        with pytest.raises(AttributeError):
            sort_order_news_arg_parser.get_str(sym_str)


@pytest.mark.unit
def test_convert_lower_symbol_str_to_correct_str():
    for sym_str, item in SortOrder.__members__.items():
        sym_str_lower = sym_str.lower()
        with pytest.raises(AttributeError):
            sort_order_news_arg_parser.get_str(sym_str_lower)


@pytest.mark.unit
def test_convert_upper_symbol_str_to_correct_str():
    for sym_str, item in SortOrder.__members__.items():
        sym_str_upper = sym_str.upper()
        with pytest.raises(AttributeError):
            sort_order_news_arg_parser.get_str(sym_str_upper)



import asyncio

import pytest

from refinitiv.data.content.ipa._ipa_content_provider import (
    IPAContentProviderLayer,
)
from refinitiv.data.errors import RDError
from tests.unit.conftest import (
    StubSession,
    error_request_failed_http_response_generator,
    error_no_location_http_response_generator,
    error_404_http_response_generator,
    success_http_response_generator,
)
from tests.unit.content.conftest import StubDefinition


@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
def test_ipa_content_provider_layer_get_data(content_type, async_mode):
    # given
    session = StubSession(is_open=True)
    session.async_mode = async_mode
    session.http_responses = success_http_response_generator()

    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    data = provider_layer.get_data(session, async_mode=async_mode)

    # then
    assert data


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
async def test_ipa_content_provider_layer_get_data_async(content_type, async_mode):
    # given
    session = StubSession(is_open=True)
    session.async_mode = async_mode
    session.http_responses = success_http_response_generator()
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    data = await provider_layer.get_data_async(session, async_mode=async_mode)

    # then
    assert data


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
async def test_ipa_content_provider_layer_get_data_pass_on_response(
    content_type, async_mode
):
    # given
    fut = asyncio.Future()

    def on_response(response, provider_layer, session):
        fut.set_result(response)

    session = StubSession(is_open=True)
    session.async_mode = async_mode
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    data = provider_layer.get_data(
        session, on_response=on_response, async_mode=async_mode
    )

    response = await fut

    # then
    assert response


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
async def test_ipa_content_provider_layer_call_get_data_async_pass_on_response(
    content_type, async_mode
):
    # given
    fut = asyncio.Future()

    def on_response(response, provider_layer, session):
        fut.set_result(response)

    session = StubSession(is_open=True)
    session.async_mode = async_mode
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    await provider_layer.get_data_async(
        session, on_response=on_response, async_mode=async_mode
    )

    response = await fut

    # then
    assert response


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
async def test_ipa_content_provider_layer_call_get_data_and_pass_on_response_session(
    content_type, async_mode
):
    # given
    fut = asyncio.Future()

    def on_response(response, provider_layer, session):
        fut.set_result(response)

    session = StubSession(is_open=True)
    session.async_mode = async_mode
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    provider_layer.get_data(session, on_response=on_response, async_mode=async_mode)

    response = await fut

    # then
    assert response


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
async def test_ipa_content_provider_layer_call_get_data_async_and_pass_on_response_session(
    content_type, async_mode
):
    # given
    fut = asyncio.Future()

    def on_response(response, provider_layer, session):
        fut.set_result(response)

    session = StubSession(is_open=True)
    session.async_mode = async_mode
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    await provider_layer.get_data_async(
        session, on_response=on_response, async_mode=async_mode
    )

    response = await fut

    # then
    assert response


@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
def test_ipa_content_provider_layer_call_get_data_and_pass_on_response_session_will_raise_error(
    content_type, async_mode
):
    # given
    def on_response(response, provider_layer, session):
        raise ValueError("exception")

    session = StubSession(is_open=True)
    session.async_mode = async_mode
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # when
    response = provider_layer.get_data(
        session, on_response=on_response, async_mode=async_mode
    )

    # then
    assert response


@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
def test_ipa_content_provider_layer_call_get_data_will_raise_error_is_session_closed(
    content_type, async_mode
):
    # given
    session = StubSession(is_open=False)
    session.async_mode = async_mode
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # then
    with pytest.raises(Exception, match="Session is not opened."):
        # when
        provider_layer.get_data(session, async_mode=async_mode)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "async_mode", [False, True], ids=["async_mode=False", "async_mode=True"]
)
async def test_ipa_content_provider_layer_call_get_data_async_will_raise_error_is_session_closed(
    content_type, async_mode
):
    # given
    session = StubSession(is_open=False)
    session.async_mode = async_mode
    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # then
    with pytest.raises(Exception, match="Session is not opened."):
        # when
        await provider_layer.get_data_async(session, async_mode=async_mode)


def test_ipa_content_provider_layer_get_data_without_202_operation_response(
    content_type,
):
    # given
    session = StubSession(is_open=True)
    session.async_mode = True
    session.http_responses = error_404_http_response_generator()

    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # then
    with pytest.raises(RDError, match="Error code 404 | Not found"):
        # when
        provider_layer.get_data(session, async_mode=True)


def test_ipa_content_provider_layer_get_data_without_location_in_operation_response(
    content_type,
):
    # given
    session = StubSession(is_open=True)
    session.async_mode = True
    session.http_responses = error_no_location_http_response_generator()

    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # then
    with pytest.raises(
        RDError,
        match="Error code None | IPA Asynchronous request operation failed, response doesn't contain location.",
    ):
        # when
        provider_layer.get_data(session, async_mode=True)


def test_ipa_content_provider_layer_get_data_wit_failed_request_response(content_type):
    # given
    session = StubSession(is_open=True)
    session.async_mode = True
    session.http_responses = error_request_failed_http_response_generator()

    provider_layer = IPAContentProviderLayer(
        content_type, definition=StubDefinition(), universe=StubDefinition()
    )

    # then
    with pytest.raises(RDError, match="Error code 400"):
        # when
        provider_layer.get_data(session, async_mode=True)



# coding: utf8


from enum import Enum, unique


@unique
class NumeraireType(Enum):
    CASH = "Cash"
    ROLLING_EVENT = "RollingEvent"
    ROLLING_PAYMENT = "RollingPayment"
    TERMINAL_EVENT_ZC = "TerminalEventZc"
    TERMINAL_ZC = "TerminalZc"



import abc
from typing import Iterable, List, Callable

from ._parsed_data import ParsedData
from ..._tools import cached_property


class BaseValidator(abc.ABC):
    @abc.abstractmethod
    def validate(self, data: ParsedData) -> bool:
        # for override
        pass


class ContentValidator(BaseValidator):
    @classmethod
    def content_data_has_no_error(cls, data: ParsedData) -> bool:
        content_data = data.content_data
        if "ErrorCode" in content_data:
            data.error_codes = content_data.get("ErrorCode")
            data.error_messages = content_data.get("ErrorMessage")
            return False

        return True

    @classmethod
    def content_data_is_not_none(cls, data: ParsedData) -> bool:
        if data.content_data is None:
            data.error_codes = 1
            data.error_messages = "Content data is None"
            return False

        return True

    @classmethod
    def content_data_status_is_not_error(cls, data: ParsedData) -> bool:
        content_data = data.content_data
        if content_data.get("status") == "Error":
            data.error_codes = content_data.get("code", -1)
            data.error_messages = content_data.get("message")
            return False

        return True

    @classmethod
    def status_is_not_error(cls, data: ParsedData) -> bool:
        if data.status == "Error":
            content_data = data.content_data
            data.error_codes = content_data.get("code")
            data.error_messages = content_data.get("message")
            return False

        return True

    @cached_property
    def validators(self) -> List[Callable[[ParsedData], bool]]:
        return [self.content_data_is_not_none, self.content_data_status_is_not_error]

    def validate(self, data: ParsedData) -> bool:
        return all(validate(data) for validate in self.validators)


class ContentTypeValidator(BaseValidator):
    def __init__(self, allowed_content_types=None):
        if allowed_content_types is None:
            allowed_content_types = {"application/json"}
        self._allowed_content_types = allowed_content_types

    def validate(self, data: ParsedData) -> bool:
        # Checking only first part (type/subtype) of media_type
        # See https://httpwg.org/specs/rfc7231.html#media.type
        content_type = (
            data.raw_response.headers.get("content-type", "").split(";")[0].strip()
        )
        is_success = content_type in self._allowed_content_types

        if not is_success:
            data.error_codes = -1
            data.error_messages = (
                f"Unexpected content-type in response,\n"
                f"Expected: {self._allowed_content_types}\n"
                f"Actual: {content_type}"
            )

        return is_success


class ValidatorContainer:
    def __init__(
        self,
        validators: Iterable = None,
        content_validator: ContentValidator = None,
        content_type_validator: ContentTypeValidator = None,
        use_default_validators=True,
    ):
        content_validator = content_validator or ContentValidator()
        content_type_validator = content_type_validator or ContentTypeValidator()

        self.validators: List[BaseValidator] = list(validators) if validators else []
        if content_type_validator and use_default_validators:
            self.validators.append(content_type_validator)
        if content_validator and use_default_validators:
            self.validators.append(content_validator)

    def validate(self, data: ParsedData) -> bool:
        return all(validator.validate(data) for validator in self.validators)



from refinitiv.data.content.estimates.view_summary import (
    annual,
    historical_snapshots_non_periodic_measures,
    historical_snapshots_periodic_measures_annual,
    historical_snapshots_periodic_measures_interim,
    historical_snapshots_recommendations,
    interim,
    non_periodic_measures,
    recommendations,
)


def test_annual(create_estimates_definition):
    annual.Definition(**create_estimates_definition)


def test_historical_snapshots_non_periodic_measures(create_estimates_definition):
    historical_snapshots_non_periodic_measures.Definition(**create_estimates_definition)


def test_historical_snapshots_periodic_measures_annual(create_estimates_definition):
    historical_snapshots_periodic_measures_annual.Definition(
        **create_estimates_definition
    )


def test_historical_snapshots_periodic_measures_interim(create_estimates_definition):
    historical_snapshots_periodic_measures_interim.Definition(
        **create_estimates_definition
    )


def test_historical_snapshots_recommendations(create_estimates_definition):
    historical_snapshots_recommendations.Definition(**create_estimates_definition)


def test_interim(create_estimates_definition):
    interim.Definition(**create_estimates_definition)


def test_non_periodic_measures(create_estimates_definition):
    non_periodic_measures.Definition(**create_estimates_definition)


def test_recommendations(create_estimates_definition):
    recommendations.Definition(**create_estimates_definition)


def test_annual_error(create_estimates_definition_with_error):
    try:
        annual.Definition(**create_estimates_definition_with_error)
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_historical_snapshots_non_periodic_measures_error(
    create_estimates_definition_with_error,
):
    try:
        historical_snapshots_non_periodic_measures.Definition(
            **create_estimates_definition_with_error
        )
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_historical_snapshots_periodic_measures_annual_error(
    create_estimates_definition_with_error,
):
    try:
        historical_snapshots_periodic_measures_annual.Definition(
            **create_estimates_definition_with_error
        )
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_historical_snapshots_periodic_measures_interim_error(
    create_estimates_definition_with_error,
):
    try:
        historical_snapshots_periodic_measures_interim.Definition(
            **create_estimates_definition_with_error
        )
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_historical_snapshots_recommendations_error(
    create_estimates_definition_with_error,
):
    try:
        historical_snapshots_recommendations.Definition(
            **create_estimates_definition_with_error
        )
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_interim_error(create_estimates_definition_with_error):
    try:
        interim.Definition(**create_estimates_definition_with_error)
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_non_periodic_measures_error(create_estimates_definition_with_error):
    try:
        non_periodic_measures.Definition(**create_estimates_definition_with_error)
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_recommendations_error(create_estimates_definition_with_error):
    try:
        recommendations.Definition(**create_estimates_definition_with_error)
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)



from enum import Enum, unique
from typing import Optional, Union

from ..._tools import make_enum_arg_parser_by_members


@unique
class CountryCode(Enum):
    """
    List of ISO 3166 country codes.

    'G:7R' is unique country code for Afghanistan
    """

    AFG = "G:7R"
    ALB = "G:7G"
    DZA = "G:7S"
    ASM = "G:39"
    AND = "G:32"
    AGO = "G:82"
    AIA = "G:1H"
    ATA = "G:AY"
    ATG = "G:31"
    ARG = "G:60"
    ARM = "G:7I"
    ABW = "G:AD"
    AUS = "G:2H"
    AUT = "G:1F"
    AZE = "G:4R"
    BHS = "G:5E"
    BHR = "G:5Q"
    BGD = "G:9B"
    BRB = "G:8P"
    BLR = "G:8B"
    BEL = "G:9Y"
    BLZ = "G:6C"
    BEN = "G:48"
    BMU = "G:75"
    BTN = "G:8F"
    BOL = "G:3J"
    BIH = "G:2D"
    BWA = "G:61"
    BVT = "G:6F"
    BRA = "G:26"
    IOT = "G:9R"
    BRN = "G:80"
    BGR = "G:1W"
    BFA = "G:5A"
    BDI = "G:68"
    KHM = "G:9G"
    CMR = "G:4I"
    CAN = "G:8W"
    CPV = "G:AF"
    CYM = "G:4J"
    CAF = "G:5G"
    TCD = "G:A6"
    CHL = "G:4M"
    CHN = "G:B1"
    CXR = "G:5R"
    CCK = "G:67"
    COL = "G:2S"
    COM = "G:99"
    COD = "G:8A"
    COG = "G:5K"
    COK = "G:1L"
    CRI = "G:5H"
    HRV = "G:5X"
    CUB = "G:7X"
    CYP = "G:8T"
    CZE = "G:2E"
    DNK = "G:19"
    DJI = "G:6Z"
    DMA = "G:8U"
    DOM = "G:76"
    ECU = "G:8Q"
    EGY = "G:3G"
    SLV = "G:AB"
    GNQ = "G:5L"
    ERI = "G:6K"
    EST = "G:9D"
    SWZ = "G:7H"
    ETH = "G:6L"
    FLK = "G:3L"
    FRO = "G:24"
    FJI = "G:3Z"
    FIN = "G:90"
    FRA = "G:5M"
    GUF = "G:3B"
    PYF = "G:54"
    ATF = "G:9V"
    GAB = "G:69"
    GMB = "G:77"
    GEO = "G:9F"
    DEU = "G:3D"
    GHA = "G:5N"
    GIB = "G:79"
    GRC = "G:6A"
    GRL = "G:2R"
    GRD = "G:9A"
    GLP = "G:4Q"
    GUM = "G:2Y"
    GTM = "G:96"
    GGY = "G:34"
    GIN = "G:9L"
    GNB = "G:9Z"
    GUY = "G:44"
    HTI = "G:22"
    HMD = "G:4W"
    HND = "G:AG"
    HKG = "G:3H"
    HUN = "G:46"
    ISL = "G:6I"
    IND = "G:5B"
    IDN = "G:25"
    IRN = "G:56"
    IRQ = "G:8G"
    IRL = "G:6X"
    IMN = "G:35"
    ISR = "G:3S"
    ITA = "G:5J"
    JAM = "G:1G"
    JPN = "G:41"
    JEY = "G:33"
    JOR = "G:1Z"
    KAZ = "G:85"
    KEN = "G:70"
    KIR = "G:7P"
    PRK = "G:AE"
    KOR = "G:83"
    KWT = "G:7Q"
    KGZ = "G:8R"
    LAO = "G:8L"
    LVA = "G:4H"
    LBN = "G:64"
    LSO = "G:2M"
    LBR = "G:3U"
    LBY = "G:6W"
    LIE = "G:A9"
    LTU = "G:8I"
    LUX = "G:7M"
    MAC = "G:3I"
    MKD = "G:AI"
    MDG = "G:7Z"
    MWI = "G:6G"
    MYS = "G:8S"
    MDV = "G:6H"
    MLI = "G:3V"
    MLT = "G:4G"
    MHL = "G:36"
    MTQ = "G:8C"
    MRT = "G:2X"
    MUS = "G:9N"
    MEX = "G:2V"
    FSM = "G:9E"
    MDA = "G:6P"
    MCO = "G:88"
    MNG = "G:66"
    MNE = "G:3E"
    MSR = "G:1X"
    MAR = "G:8X"
    MOZ = "G:2B"
    MMR = "G:72"
    NAM = "G:6Q"
    NRU = "G:8J"
    NPL = "G:2J"
    NLD = "G:7K"
    NCL = "G:2L"
    NZL = "G:49"
    NIC = "G:AC"
    NER = "G:2U"
    NGA = "G:6B"
    NIU = "G:62"
    NFK = "G:7Y"
    NOR = "G:3N"
    OMN = "G:7B"
    PAK = "G:2P"
    PLW = "G:2N"
    PSE = "G:59"
    PAN = "G:4U"
    PNG = "G:2G"
    PRY = "G:89"
    PER = "G:3T"
    PHL = "G:7L"
    PCN = "G:15"
    POL = "G:5Y"
    PRT = "G:A3"
    PRI = "G:5U"
    QAT = "G:51"
    REU = "G:6N"
    ROU = "G:2Z"
    RUS = "G:38"
    RWA = "G:AA"
    SHN = "G:9S"
    KNA = "G:40"
    LCA = "G:3A"
    SPM = "G:4E"
    VCT = "G:3F"
    WSM = "G:2F"
    SMR = "G:78"
    STP = "G:5F"
    SAU = "G:92"
    SEN = "G:6E"
    SRB = "G:7F"
    SYC = "G:5C"
    SLE = "G:A5"
    SGP = "G:7D"
    SVK = "G:1C"
    SVN = "G:74"
    SLB = "G:1Y"
    SOM = "G:5D"
    ZAF = "G:2I"
    SGS = "G:1N"
    SSD = "G:C2"
    ESP = "G:55"
    LKA = "G:1J"
    SDN = "G:C1"
    SUR = "G:86"
    SJM = "G:1M"
    SWE = "G:6V"
    CHE = "G:30"
    SYR = "G:4P"
    TWN = "G:7U"
    TJK = "G:4N"
    TZA = "G:2T"
    THA = "G:3R"
    TGO = "G:91"
    TKL = "G:5P"
    TON = "G:8K"
    TTO = "G:9T"
    TUN = "G:2W"
    TUR = "G:8Z"
    TKM = "G:42"
    TCA = "G:9I"
    TUV = "G:2C"
    UGA = "G:47"
    UKR = "G:71"
    ARE = "G:A4"
    GBR = "G:7J"
    UMI = "G:9W"
    USA = "G:6J"
    URY = "G:4Y"
    UZB = "G:8M"
    VAT = "G:8Y"
    VUT = "G:9M"
    VEN = "G:2K"
    VNM = "G:5Z"
    WLF = "G:4L"
    ESH = "G:4F"
    YEM = "G:28"
    ZMB = "G:73"
    ZWE = "G:52"


OptCountryCode = Optional[Union[str, CountryCode]]

country_code_arg_parser = make_enum_arg_parser_by_members(CountryCode)



import os

import allure
import pytest

import refinitiv.data as rd
from refinitiv.data import _configure as configure
from tests.integration.configure.conftest import open_session_and_get_base_url
from tests.integration.helpers import compare_list

BASE_URL_BETA = "https://api.ppe.refinitiv.com"
BASE_URL_PROD = "https://api.refinitiv.com"
ENV_NAME = os.environ.get("ENV_NAME")


@allure.suite("Configuration file")
@allure.feature("Configuration file")
@allure.severity(allure.severity_level.CRITICAL)
class TestConfigFile:
    @allure.title("Check default base url for desktop session")
    @pytest.mark.caseid("37655144")
    def test_default_base_url_for_desktop_session(self, open_desktop_session):
        session = open_desktop_session
        base_url = session._get_base_url()
        expected_url = ["http://localhost:9000", "http://localhost:9001", "http://localhost:9060"]
        assert base_url in expected_url, f"Actual url: {base_url}"

    @allure.title("Check default base url for platform session")
    @pytest.mark.caseid("37655149")
    def test_default_base_url_for_platform_session(self, open_platform_session):
        session = open_platform_session
        base_url = session._base_url

        expected_url = None
        if ENV_NAME == "PROD":
            expected_url = BASE_URL_PROD
        elif ENV_NAME == "BETA":
            expected_url = BASE_URL_BETA

        assert expected_url == base_url, f"Actual url: {base_url}"

    @allure.title("Override default pricing endpoint")
    @pytest.mark.caseid("37655231")
    def test_override_default_pricing_endpoint(self):
        test_endpoint = "/test"
        rd.get_config().set_param(
            param=f"apis.data.pricing.endpoints", value=test_endpoint
        )
        actual_endpoint = rd.get_config().get_param(
            param=f"apis.data.pricing.endpoints"
        )

        assert actual_endpoint == test_endpoint, f"{actual_endpoint} != {test_endpoint}"

    @allure.title("Check default ownership Fund Top-n-Concentration url")
    @pytest.mark.caseid("37655273")
    def test_default_ownership_fund_top_n_concentration_url(
        self, open_platform_session
    ):
        response = rd.content.ownership.fund.top_n_concentration.Definition(
            universe="ONEX.CCP", count=5
        ).get_data()
        url = response.request_message.url.path
        expected_url = rd.get_config().get_param(
            param=f"apis.data.ownership.endpoints.fund.top-n-concentration"
        )
        assert expected_url in url, f"Actual url: {url}"

    @allure.title("Check keys from config file at first level")
    @pytest.mark.caseid("37655346")
    def test_keys_from_config_file_on_first_level(self):
        configure.reload()
        actual_keys = rd.get_config().keys(1)
        expected_keys = [
            "raise_exception_on_error",
            "apis",
            "http",
            "config-change-notifications-enabled",
            "sessions",
            "logs",
            "bulk",
            'usage_logger'
        ]

        compare_list(actual_keys, expected_keys)

    @allure.title("Open two sessions and load different config file for every session")
    @pytest.mark.caseid("37691516")
    def test_open_two_session_and_load_diff_config_file_for_every_session(self):
        rd.load_config("./tests/integration/configure/beta-config.json")
        session = rd.session.platform.Definition(name="my-session").get_session()
        base_url = open_session_and_get_base_url(session)
        expected_url = BASE_URL_BETA

        assert expected_url == base_url, f"Actual url: {base_url}"

        rd.load_config("./tests/integration/configure/refinitiv-data.config.json")
        session = rd.session.platform.Definition(name="my-test-session").get_session()
        base_url = open_session_and_get_base_url(session)
        expected_url = BASE_URL_PROD

        assert expected_url == base_url, f"Actual url: {base_url}"



import allure
import pytest

from refinitiv.data.content import search
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.search.conftest import (
    check_document_title_contains_query,
    check_business_entity_for_every_document_in_response,
    check_navigator_entity_in_response,
    check_hits_and_dataframe_columns,
)
from tests.integration.helpers import (
    check_response_status,
    check_non_empty_response_data,
    check_extended_params_were_sent_in_request,
    get_async_response_from_definitions,
)


@allure.suite("Content object - Search")
@allure.feature("Content object - Search")
@allure.severity(allure.severity_level.CRITICAL)
class TestSearch:
    @allure.title(
        "Create a Content object Search with query using a string and get data"
    )
    @pytest.mark.caseid("35382768")
    @pytest.mark.parametrize("query", ["Microsoft"])
    def test_create_content_object_search_with_query_using_string_and_get_data(
        self, open_session, query
    ):
        response = search.Definition(query).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_document_title_contains_query(response, query)

    @allure.title("Create a Content object Search when session is not opened")
    @pytest.mark.caseid("35384395")
    @pytest.mark.parametrize("query", ["Microsoft"])
    def test_create_content_object_search_when_session_is_not_opened(self, query):
        with pytest.raises(AttributeError) as error:
            search.Definition(query).get_data()
        assert str(error.value) == "No default session created yet. Please create a session first!"

    @allure.title("Create a Content object Search with query and optional parameters")
    @pytest.mark.caseid("35385811")
    @pytest.mark.parametrize("query", ["Tesla"])
    def test_create_content_object_search_with_query_and_optional_parameters(
        self, open_session, query
    ):
        response = search.Definition(query, view=search.Views.PEOPLE).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_business_entity_for_every_document_in_response(response, "PERSON")

    @allure.title("Create a Content object Search with invalid query")
    @pytest.mark.caseid("35385844")
    @pytest.mark.parametrize("query", [111])
    def test_create_content_object_search_with_invalid_query(self, open_session, query):
        with pytest.raises(RDError) as error:
            search.Definition(query).get_data()
        assert (
            str(error.value)
            == "Error code 400 | Validation error: json.Query in body must be of type string"
        )

    @allure.title(
        "Create a Content object Search with valid query and invalid optional parameters"
    )
    @pytest.mark.caseid("35385858")
    @pytest.mark.parametrize(
        "query,expected_error",
        [
            (
                "IBM",
                "Error code 400 | Validation error: json.View in body should be one of",
            )
        ],
    )
    def test_create_content_object_search_with_valid_query_and_invalid_optional_parameters(
        self, open_session, query, expected_error
    ):
        definition = search.Definition(query, view="invalid")
        with pytest.raises(RDError) as error:
            definition.get_data()
        assert expected_error in str(error.value)

    @allure.title("Extended params: override all fields")
    @pytest.mark.caseid("35513512")
    @pytest.mark.smoke
    @pytest.mark.parametrize("query,overridden_query", [("Microsoft", "Tesla")])
    def test_create_content_object_search_with_extended_params_overriding_all_fields(
        self, open_session, query, overridden_query
    ):
        extended_params = {"Query": overridden_query, "View": "Organisations"}
        response = search.Definition(
            query, view=search.Views.PEOPLE, extended_params=extended_params
        ).get_data()
        check_extended_params_were_sent_in_request(response, extended_params)
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_document_title_contains_query(response, overridden_query)

    @allure.title("Extended params: send additional fields")
    @pytest.mark.caseid("36105757")
    @pytest.mark.parametrize("query", ["IBM"])
    def test_create_content_object_search_with_extended_params_with_additional_fields(
        self, open_session, query
    ):
        extended_params = {
            "View": "Organisations",
            "INVALID_PARAM": "INVALID_PARAM_VALUE",
        }
        response = search.Definition(query, extended_params=extended_params).get_data()
        check_extended_params_were_sent_in_request(response, extended_params)
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)

    @allure.title(
        "Create a Content object Search with valid and invalid definition get data asynchronously"
    )
    @pytest.mark.caseid("35382768")
    @pytest.mark.parametrize(
        "query_1,query_2,order_by,select",
        [
            (
                "Oracle",
                "ceo",
                ["YearOfBirth desc,LastName,FirstName"],
                "YearOfBirth,DocumentTitle",
            )
        ],
    )
    async def test_create_content_object_search_with_query_using_string_and_get_data_asynchronously(
        self, open_session_async, query_1, query_2, order_by, select
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            search.Definition(query=query_1),
            search.Definition(query=query_2, order_by=order_by, select=select),
        )

        check_response_status(valid_response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(valid_response)
        check_document_title_contains_query(valid_response, query_1)

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED,
            expected_http_reason=HttpReason.BAD_REQUEST,
            expected_error_message='Validation error: json.OrderBy in body must be of type string: "array"',
            expected_error_code=HttpStatusCode.FOUR_HUNDRED,
        )

    @allure.title("Create a Content object Search with navigators")
    @pytest.mark.caseid("C51600314")
    @pytest.mark.parametrize("navigator", [("Currency")])
    def test_create_content_object_search_with_navigators(
        self, open_session, navigator
    ):
        response = search.Definition(
            view=search.Views.GOV_CORP_INSTRUMENTS,
            top=0,
            navigators=f"{navigator}(buckets:10,desc:sum_FaceOutstandingUSD,calc:max_CouponRate)",
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_navigator_entity_in_response(
            response,
            navigator,
            navigator_props=["sum_FaceOutstandingUSD", "max_CouponRate"],
        )

        assert response.data.df.empty, f"Empty response.data.df received"
        assert len(response.data.hits) == 0

    @allure.title("Create a Content object Search with sub navigators")
    @pytest.mark.caseid("C51600315")
    @pytest.mark.parametrize(
        "hits_item_attributes,navigator_1,sub_navigator_1,navigator_2,sub_navigator_2",
        [
            (
                ["BusinessEntity", "DocumentTitle", "PI", "PermID", "RIC"],
                "MaturityDate",
                "BusinessEntity",
                "FirstName",
                "LastName",
            )
        ],
    )
    def test_create_content_object_search_with_sub_navigators(
        self,
        open_session,
        hits_item_attributes,
        navigator_1,
        sub_navigator_1,
        navigator_2,
        sub_navigator_2,
    ):
        response = search.Definition(
            view="GovCorpInstruments",
            filter="DbType eq 'CORP' and IsActive eq true and (MaturityDate ge 2020-01-01 and MaturityDate le 2025-12-31)",
            navigators=f"{navigator_1}(type:histogram, buckets:7, sub:{sub_navigator_1}(calc:sum_FaceOutstandingUSD)),"
            f"{navigator_2}(buckets:5,sub:{sub_navigator_2}(buckets:2))",
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_hits_and_dataframe_columns(response, hits_item_attributes)
        check_navigator_entity_in_response(response, navigator_1, sub_navigator_1)
        check_navigator_entity_in_response(response, navigator_2, sub_navigator_2)



import pytest

from refinitiv.data.content.ipa.financial_contracts import cross
from tests.unit.conftest import StubSession


@pytest.fixture(scope="function")
def cross_stream():
    session = StubSession(is_open=True)
    stream = cross.Definition(
        instrument_tag="00102700008910C",
        fx_cross_type=cross.FxCrossType.FX_FORWARD,
        fx_cross_code="USDEUR",
        legs=[
            cross.LegDefinition(
                end_date="2015-04-09T00:00:00Z",
            )
        ],
        pricing_parameters=cross.PricingParameters(
            valuation_date="2015-02-02T00:00:00Z",
            price_side=cross.PriceSide.MID,
        ),
        fields=[
            "InstrumentTag",
            "InstrumentDescription",
            "ValuationDate",
            "FxOutrightCcy1Ccy2",
        ],
    ).get_stream(session=session)
    return stream



from ema import ElementList, ReqMsg, ENAME_VIEW_DATA, ENAME_VIEW_TYPE, VT_FIELD_ID_LIST

from .ema import IntArray


def json_marketprice_msg_to_ema(msg, rdm_dict):
    streaming = msg.get("Streaming", True)
    service_name = msg["Key"].get("Service", "ELEKTRON_DD")
    name = msg["Key"]["Name"]

    ema_msg = (
        ReqMsg().name(name).service_name(service_name).interest_after_refresh(streaming)
    )

    if msg.get("View"):
        # WARNING: Non-existent fields will be skipped
        field_ids = [rdm_dict[item] for item in msg["View"] if item in rdm_dict]
        view_data = IntArray(field_ids).complete()
        ema_msg.payload(
            ElementList()
            .add_uint(ENAME_VIEW_TYPE, VT_FIELD_ID_LIST)
            .add_array(ENAME_VIEW_DATA, view_data)
            .complete()
        )
    return ema_msg



from dataclasses import dataclass, field
from threading import Event
from typing import Tuple, List, TYPE_CHECKING

from ._universe_expander import UniverseExpander
from ..._core.session import get_default
from ..._errors import RDError, ScopeError
from ..._tools import cached_property
from ...content import fundamental_and_reference
from ...content.pricing import chain
from ...content.pricing.chain._stream_facade import Stream
from ...delivery import endpoint_request
from ...delivery._data._data_provider_layer import _check_response


if TYPE_CHECKING:
    from ...delivery._data._response import Response

default_error_message = "No values to unpack"


def on_error_callback(data: Tuple[dict], universe: str, stream: Stream):
    state = data[0].get("State", {})
    message = state.get("Text", default_error_message)
    code = state.get("Code", -1)
    stream.close()
    raise RDError(code, f"{message}\nuniverse: {universe}")


def get_chain_data_from_stream(name):
    constituents, summary_links = [], []
    is_stream_closed = Event()

    def on_complete(_, stream):
        nonlocal constituents, summary_links, is_stream_closed
        stream.close()
        constituents, summary_links = stream.constituents, stream.summary_links
        is_stream_closed.set()

    chain_stream = chain.Definition(name).get_stream()
    chain_stream.on_complete(on_complete)
    chain_stream.on_error(on_error_callback)
    chain_stream.open()
    is_stream_closed.wait()
    if not constituents:
        raise RDError(-1, default_error_message)
    return ChainData(constituents, summary_links)


def get_chain_data_from_adc(name):
    adc_response = fundamental_and_reference.Definition(
        universe=name, fields=["TR.RIC"]
    ).get_data()
    constituents = [item[0] for item in adc_response.data.raw.get("data", {})]
    return ChainData(constituents)


def _get_constituents(response: "Response") -> list:
    return response.data.raw.get("data", {}).get("constituents", [])


def get_chain_data_from_chain_endpoint(name):
    url = "/data/pricing/chains/v1/"
    session = get_default()
    session.verify_scope(url, "get")
    chain_response = endpoint_request.Definition(
        url=url, query_parameters={"universe": name}
    ).get_data()
    _check_response(chain_response, session.config)

    summary_links = []
    constituents = []

    for item in _get_constituents(chain_response):
        if item.startswith(".") or item.startswith("/") or item.endswith("="):
            summary_links.append(item)
        else:
            constituents.append(item)

    next_link = chain_response.data.raw.get("meta", {}).get("nextLink")
    while next_link:
        chain_response = endpoint_request.Definition(
            url=url, query_parameters={"universe": name, "target": next_link}
        ).get_data()

        next_link = chain_response.data.raw.get("meta", {}).get("nextLink")
        constituents.extend(_get_constituents(chain_response))

    return ChainData(constituents, summary_links)


def get_chain_data(name):
    for func in (
        get_chain_data_from_stream,
        get_chain_data_from_adc,
    ):
        try:
            return func(name)
        except ScopeError:
            continue
    return get_chain_data_from_chain_endpoint(name)


@dataclass
class ChainData:
    constituents: List = field(default_factory=lambda: [])
    summary_links: List = field(default_factory=lambda: [])


class Chain(UniverseExpander):
    """
    Class to get data from chain.

    Parameters
    ----------
    name : str
        chain name


    Examples
    --------
    >>> chain = Chain("0#.DJI")
    >>> print(list(chain))
    >>> print(chain.constituents)
    >>> print(chain.summary_links)
    """

    def __init__(self, name):
        self._name = name

    @cached_property
    def _chains(self):
        return get_chain_data(self._name)

    @property
    def name(self):
        return self._name

    @property
    def summary_links(self):
        return self._chains.summary_links

    @property
    def constituents(self):
        return self._chains.constituents

    @property
    def _universe(self):
        return self.constituents



from typing import TYPE_CHECKING, List

from .._content_data_validator import ContentDataValidator
from .._request_factory import DatesAndCalendarsResponseFactory
from ..holidays._holidays_data_provider import Holiday
from ...._content_data import Data
from ...._content_data_provider import ContentDataProvider
from ...._df_builder import build_dates_calendars_df
from .....content.ipa._content_provider import DatesAndCalendarsRequestFactory
from .....delivery._data._data_provider import ValidatorContainer

if TYPE_CHECKING:
    from .....delivery._data._data_provider import ParsedData


class Period:
    _holidays = None

    def __init__(self, date: str, holidays: list = None, tag: str = ""):
        self._date = date
        self._response_holidays_items = holidays or []
        self._tag = tag

    @property
    def tag(self):
        return self._tag

    @property
    def date(self):
        return self._date

    @property
    def holidays(self):
        if self._holidays is None:
            self._holidays = [
                Holiday(holiday=holiday, tag=self.tag)
                for holiday in self._response_holidays_items
            ]

        return self._holidays


class AddedPeriods(Data):
    _added_periods = None

    @property
    def added_periods(self):
        if self._added_periods is None:
            self._added_periods = [
                Period(
                    date=raw_item["date"],
                    holidays=raw_item.get("holidays"),
                    tag=raw_item.get("tag"),
                )
                for raw_item in self._raw
                if not raw_item.get("error")
            ]

        return self._added_periods

    def __getitem__(self, item):
        return self.added_periods[item]


class AddedPeriod(Data):
    def __init__(
        self, raw: List[dict], date: str, holidays: None, tag: str = "", **kwargs
    ):
        super().__init__(raw, **kwargs)
        self._period = Period(date=date, holidays=holidays, tag=tag)

    @property
    def added_period(self):
        return self._period


class AddPeriodsResponseFactory(DatesAndCalendarsResponseFactory):
    def create_data_success(self, parsed_data: "ParsedData", **kwargs):
        raw: List[dict] = self.get_raw(parsed_data)

        if len(raw) > 1:
            data = AddedPeriods(raw, build_dates_calendars_df)

        else:
            raw_item = raw[0]
            data = AddedPeriod(
                raw=raw,
                date=raw_item["date"],
                holidays=raw_item.get("holidays"),
                tag=raw_item.get("tag"),
                dfbuilder=build_dates_calendars_df,
            )

        return data


add_period_data_provider = ContentDataProvider(
    request=DatesAndCalendarsRequestFactory(),
    response=AddPeriodsResponseFactory(),
    validator=ValidatorContainer(content_validator=ContentDataValidator()),
)



from typing import TYPE_CHECKING, Any, Union, Dict
from urllib.parse import quote_plus

from pandas import DataFrame

from .._tools import (
    ADC_FUNC_PATTERN_IN_FIELDS,
    ADC_TR_PATTERN,
    fields_arg_parser,
    cached_property,
    iterator_str_arg_parser,
)

if TYPE_CHECKING:
    from .._types import OptStrStrs, StrStrings


class Container:
    def __init__(self, raw: Any = None) -> None:
        self._raw = raw

    @property
    def raw(self) -> Any:
        return self._raw

    def __bool__(self) -> bool:
        return bool(self._raw)

    def __iter__(self):
        return iter(self._raw)


class UniverseContainer(Container):
    def __init__(self, raw: "OptStrStrs" = None) -> None:
        super().__init__(raw)
        self._hp = []

    @cached_property
    def _universe(self) -> "StrStrings":
        return iterator_str_arg_parser.get_list(self._raw)

    @cached_property
    def hp_and_cust_inst(self):
        return self.hp + self.cust_inst

    @cached_property
    def adc(self) -> "StrStrings":
        return [inst for inst in self._universe if not inst.startswith("S)")]

    @cached_property
    def cust_inst(self) -> "StrStrings":
        return [inst for inst in self._universe if inst.startswith("S)")]

    @property
    def hp(self) -> "StrStrings":
        return self._hp

    @cached_property
    def is_universe_expander(self):
        from ..discovery._universe_expanders._universe_expander import UniverseExpander

        return isinstance(self._raw, UniverseExpander)

    def calc_hp(self, adc_raw: Union[Dict, None]) -> None:
        if adc_raw is not None:
            adc_rics = [
                quote_plus(item[0]) if "/" in item[0] else item[0]
                for item in adc_raw.get("data", {})
            ]
        else:
            adc_rics = None
        self._hp = adc_rics or self.adc

    def __iter__(self):
        return iter(self._universe)

    def __len__(self):
        return len(self._universe)


class FieldsContainer(Container):
    def __init__(self, raw: "OptStrStrs" = None) -> None:
        super().__init__(raw)
        self._adc_fields: "OptStrStrs" = None
        self._hp_fields: "OptStrStrs" = None

    def _parse(self) -> None:
        self._adc_fields = []
        self._hp_fields = []

        for field in self._fields:
            if ADC_TR_PATTERN.match(field) or ADC_FUNC_PATTERN_IN_FIELDS.match(field):
                self._adc_fields.append(field)
            else:
                self._hp_fields.append(field)

    @cached_property
    def _fields(self) -> "StrStrings":
        return fields_arg_parser.get_list(self._raw or [])

    @property
    def adc(self) -> "StrStrings":
        if self._adc_fields is None:
            self._parse()
        return self._adc_fields

    @property
    def hp(self) -> "StrStrings":
        if self._hp_fields is None:
            self._parse()
        return self._hp_fields

    @property
    def is_no_hp(self) -> bool:
        return not self.hp

    @cached_property
    def is_disjoint_adc(self) -> bool:
        return set(self.adc).isdisjoint(set(self._fields))

    @cached_property
    def is_one_adc_no_hp(self) -> bool:
        return len(self.adc) == 1 and not self.hp

    def __getattr__(self, attr: str) -> Any:
        try:
            return getattr(self._fields, attr)
        except KeyError:
            raise AttributeError(attr)

    def __iter__(self):
        return iter(self._fields)


class DataContainer(Container):
    def __init__(
        self, raw: Union[dict, list, None], df: Union[DataFrame, None]
    ) -> None:
        super().__init__(raw)
        self._df = df

    @property
    def df(self) -> DataFrame:
        return self._df


class ADCDataContainer(DataContainer):
    def __init__(
        self,
        raw: Union[dict, list, None],
        df: Union[DataFrame, None],
        fields: "FieldsContainer",
    ):
        super().__init__(raw, df)
        self._fields = fields

    def __bool__(self):
        is_none = self._raw in [{}, None] or (
            self._raw and self._fields.is_disjoint_adc
        )
        return not is_none


class HPDataContainer(DataContainer):
    pass


class CustInstDataContainer(DataContainer):
    pass


class HPAndCustInstDataContainer(HPDataContainer):
    def __init__(
        self,
        columns: Union[list, None],
        raw: Union[dict, list, None],
        df: Union[DataFrame, None],
    ):
        super().__init__(raw, df)
        self._columns = columns

    @property
    def columns(self):
        return self._columns



__all__ = (
    "AssetClass",
    "CalendarAdjustment",
    "CompoundingType",
    "ConvexityAdjustment",
    "DayCountBasis",
    "Definition",
    "Definitions",
    "ExtrapolationMode",
    "ForwardCurveDefinition",
    "InterpolationMode",
    "PriceSide",
    "RiskType",
    "Step",
    "SwapZcCurveDefinition",
    "SwapZcCurveParameters",
    "Turn",
    "Outputs",
)

from ._definition import Definition, Definitions
from ..._curves._enums import ForwardCurvesOutputs as Outputs
from refinitiv.data.content.ipa.curves.forward_curves._models import (
    ConvexityAdjustment,
    Step,
    Turn,
)
from ..._curves import (
    ForwardCurveDefinition,
    SwapZcCurveDefinition,
    SwapZcCurveParameters,
)

from ._enums import (
    AssetClass,
    RiskType,
    DayCountBasis,
    InterpolationMode,
    PriceSide,
    ExtrapolationMode,
    CalendarAdjustment,
    CompoundingType,
)



__all__ = (
    "_check_response",
    "BaseResponse",
    "ContentTypeValidator",
    "ContentValidator",
    "DataProvider",
    "DataProviderLayer",
    "emit_event",
    "EndpointData",
    "Error",
    "ParsedData",
    "Parser",
    "Request",
    "RequestFactory",
    "Response",
    "ResponseFactory",
    "success_http_codes",
    "ValidatorContainer",
)

from typing import TYPE_CHECKING, Union

from ._connection import HttpSessionConnection
from ._data_provider_layer import DataProviderLayer, emit_event, _check_response
from ._endpoint_data import EndpointData, Error
from ._parsed_data import ParsedData
from ._raw_data_parser import Parser, success_http_codes
from ._request import Request
from ._request_factory import RequestFactory
from ._response import BaseResponse, Response
from ._response_factory import ResponseFactory
from ._validators import (
    ValidatorContainer,
    ContentValidator,
    ContentTypeValidator,
    BaseValidator,
)
from ..._core.session import raise_if_closed

if TYPE_CHECKING:
    import httpx


class DataProvider:
    def __init__(
        self,
        connection: HttpSessionConnection = None,
        request: RequestFactory = None,
        response: ResponseFactory = None,
        parser: Parser = None,
        validator: Union[BaseValidator, ValidatorContainer] = None,
    ):
        super().__init__()
        self.connection = connection or HttpSessionConnection()
        self.request = request or RequestFactory()
        self.response = response or ResponseFactory()
        self.parser = parser or Parser()
        self.validator = validator or ValidatorContainer()

    def _process_response(
        self, raw_response: "httpx.Response", session, *args, **kwargs
    ) -> BaseResponse:
        is_success, parsed_data = self.parser.parse_raw_response(raw_response)
        is_success = is_success and self.validator.validate(parsed_data)
        return self.response.create_response(
            is_success, parsed_data=parsed_data, session=session, **kwargs
        )

    def get_data(self, session, *args, **kwargs) -> BaseResponse:
        raise_if_closed(session)
        request = self.request.create(session, *args, **kwargs)
        raw_response = self.connection.send(request, session, *args, **kwargs)
        return self._process_response(raw_response, session, *args, **kwargs)

    async def get_data_async(self, session, *args, **kwargs) -> BaseResponse:
        raise_if_closed(session)
        request = self.request.create(session, *args, **kwargs)
        raw_response = await self.connection.send_async(
            request, session, *args, **kwargs
        )
        return self._process_response(raw_response, session, *args, **kwargs)


default_data_provider = DataProvider()



import allure
import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.errors import RDError
from tests.integration.content.ipa.contracts.cds.conftest import (
    cds_definition,
    cds_universe,
)
from tests.integration.content.ipa.contracts.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
    check_stream_state_and_df_from_stream,
    add_call_backs,
)
from tests.integration.helpers import (
    get_async_response_from_definition,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - CDS")
@allure.feature("Content object - CDS")
@allure.severity(allure.severity_level.CRITICAL)
class TestCDS:
    @pytest.mark.xfail(
        reason="Server side problem, time to time it raise technical error"
    )
    @allure.title("Create a cds definition object with params and get data")
    @pytest.mark.caseid("36523217")
    def test_cds_definition_get_data(self, open_session):
        response = cds_definition().get_data()
        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create a cds definition without any parameters")
    @pytest.mark.caseid("36523223")
    def test_get_cds_analytics_without_any_parameters(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.cds.Definition().get_data()
        assert (
            str(error.value)
            == "Error code QPS-DPS.1012 | Invalid input: leg definition is invalid. The Premium leg is empty."
        )

    @allure.title("Create a cds definition with empty pricing parameters")
    @pytest.mark.caseid("36523226")
    def test_get_cds_with_empty_pricing_params(self, open_session):
        response = rdf.cds.Definition(
            instrument_tag="Cds1_InstrumentCode",
            instrument_code="BNPP5YEUAM=R",
            cds_convention=rdf.cds.CdsConvention.ISDA,
            end_date_moving_convention=rdf.cds.BusinessDayConvention.NO_MOVING,
            adjust_to_isda_end_date=True,
            pricing_parameters=rdf.cds.PricingParameters(),
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a cds definition with invalid extended parameters")
    @pytest.mark.caseid("36523459")
    def test_get_cds_with_invalid_extended_params(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.cds.Definition(
                instrument_tag="Cds1_InstrumentCode",
                instrument_code="BNPP5YEUAM=R",
                cds_convention=rdf.cds.CdsConvention.ISDA,
                end_date_moving_convention=rdf.cds.BusinessDayConvention.NO_MOVING,
                adjust_to_isda_end_date=True,
                fields=[
                    "InstrumentTag",
                    "ValuationDate",
                    "InstrumentDescription",
                    "StartDate",
                    "EndDate",
                ],
                extended_params={"instrumentDefinition": "INVALID"},
            ).get_data()
        assert (
            str(error.value)
            == "Error code 400 | Invalid input: Unbindable json. Error: Could not cast or convert from System.String to TR.Qps.Calculators.BL.CdS.DataModel.CdsDefinition. Path: instrumentDefinition"
        )

    @allure.title("Create a cds definition request asynchronously")
    @pytest.mark.caseid("36523460")
    async def test_cds_definition_asynchronously(self, open_session_async):
        response = await get_async_response_from_definition(cds_definition())
        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create a cds stream")
    @pytest.mark.caseid("36523463")
    def test_cds_stream(self, open_session_async):
        stream = cds_definition().get_stream()
        stream.open()

        check_stream_state_and_df_from_stream(stream)

    @allure.title(
        "Create a cds stream with callback on_response and extended parameters"
    )
    @pytest.mark.caseid("36524104")
    def test_cds_stream_with_callback_and_extended_params(self, open_session):
        events_list = []
        stream = rdf.cds.Definition(
            instrument_tag="Cds1_InstrumentCode",
            instrument_code="BNPP5YEUAM=R",
            end_date_moving_convention=rdf.cds.BusinessDayConvention.NEXT_BUSINESS_DAY,
            pricing_parameters=rdf.cds.PricingParameters(market_data_date="2020-01-01"),
            fields=[
                "InstrumentTag",
                "ValuationDate",
                "InstrumentDescription",
                "StartDate",
                "EndDate",
                "InstrumentTag",
                "ValuationDate",
                "MarketDataDate",
                "ErrorMessage",
            ],
            extended_params={
                "instrumentDefinition": {
                    "cdsConvention": "ISDA",
                    "endDateMovingConvention": "NoMoving",
                    "adjustToIsdaEndDate": True,
                },
            },
        ).get_stream(open_session)

        add_call_backs(stream, events_list)
        stream.open()

        check_stream_state_and_df_from_stream(stream)
        assert (
            stream._stream.universe == cds_universe
        ), f"Extended params are applied incorrectly"
        assert "Response received for" in events_list, f"Events list is {events_list}"



# coding: utf8
from threading import Event
from typing import Optional, Any
from typing import TYPE_CHECKING

from ._protocol_type import ProtocolType
from ._stream_cxn_cache import stream_cxn_cache
from .event import StreamEvent, StreamCxnEvent
from .stream_state import StreamState
from .stream_state_manager import StreamStateManager
from ..._core.session.tools import is_closed
from ..._errors import ScopeError

if TYPE_CHECKING:
    from ..._content_type import ContentType
    from .._stream import StreamConnection
    from ..._core.session import Session
    from ._stream_factory import StreamDetails


class Stream(StreamStateManager):
    _cxn: Optional["StreamConnection"] = None
    _event: Optional[StreamEvent] = None

    def __init__(
        self,
        stream_id: int,
        session: "Session",
        details: "StreamDetails",
    ) -> None:
        StreamStateManager.__init__(self, logger=session.logger())
        self.details = details
        self._id: int = stream_id
        self._session: "Session" = session
        self._opened: Optional[Event] = Event()
        self._classname = f"[{self.__class__.__name__}_{self.id}]"

    @property
    def classname(self):
        return self._classname

    @property
    def id(self) -> int:
        return self._id

    @property
    def session(self) -> "Session":
        return self._session

    @session.setter
    def session(self, session: "Session"):
        if self._session != session and not self.is_open:
            self._session = session
            StreamStateManager.__init__(self, logger=self._session.logger())

    @property
    def name(self) -> str:
        return ""

    @property
    def content_type(self) -> "ContentType":
        return self.details.content_type

    @property
    def protocol_type(self) -> ProtocolType:
        return ProtocolType.NONE

    @property
    def close_message(self) -> dict:
        return {}

    @property
    def open_message(self) -> dict:
        return {}

    def open(self, *args, with_updates: bool = True) -> StreamState:
        if is_closed(self._session):
            raise AssertionError("Session must be open")

        return super().open(with_updates=with_updates)

    def send(self, message) -> bool:
        if self._cxn:
            self._debug(f"{self._classname} send {message}")
            return self._cxn.send_message(message)
        else:
            self._debug(
                f"{self._classname} cannot send {message}, "
                f"cxn is {self._cxn}, state is {self.state}"
            )
            return False

    def _initialize_cxn(self):
        self._event = StreamEvent.get(self.id)
        try:
            self._cxn = stream_cxn_cache.get_cxn(self.session, self.details)
        except ScopeError as e:
            self.halt()
            raise e
        self._cxn.on(StreamCxnEvent.DISCONNECTING, self.close)
        self._cxn.on(StreamCxnEvent.RECONNECTED, self._on_reconnected)
        self._cxn.on(StreamCxnEvent.DISPOSED, self.halt)

    def _release_cxn(self):
        self._cxn.remove_listener(StreamCxnEvent.DISCONNECTING, self.close)
        self._cxn.remove_listener(StreamCxnEvent.RECONNECTED, self._on_reconnected)
        self._cxn.remove_listener(StreamCxnEvent.DISPOSED, self.halt)
        self._debug(f"{self._classname} release cxn={self._cxn}")
        if stream_cxn_cache.has_cxn(self.session, self.details):
            stream_cxn_cache.release(self.session, self.details)
        self._cxn = None

    def _do_open(self, *args, **kwargs) -> None:
        self._opened.clear()
        self.send(self.open_message)
        if not self._cxn.is_disposed:
            self._opened.wait()

    def _do_close(self, *args, **kwargs):
        self.send(self.close_message)
        self._dispose()

    def _on_reconnected(self, *args, **kwargs):
        if self.is_open:
            self.send(self.open_message)

    def _do_on_stream_error(self, originator, *args) -> Any:
        if self.is_opening:
            self._opened.set()

        return args


def update_message_with_extended_params(message: dict, extended_params: dict) -> dict:
    return update_key_in_dict(message, extended_params)


def update_key_in_dict(message: dict, extended_params: dict) -> dict:
    for param, extended_val in extended_params.items():
        if param in message:
            prev_value = message[param]
            if isinstance(prev_value, dict) and isinstance(extended_val, dict):
                update_key_in_dict(prev_value, extended_val)
            else:
                message[param] = extended_val
        else:
            message[param] = extended_val

    return message



import json

import allure
import pytest

from refinitiv.data.content.ipa.curves import forward_curves
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.curves.conftest import (
    forward_curve_definition_01,
    forward_curve_definition_02,
    check_curve,
    check_curve_response,
)
from tests.integration.content.ipa.curves.forward_curve.conftest import forward_curve_universe
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - ForwardCurve")
@allure.feature("Content object - ForwardCurve")
@allure.severity(allure.severity_level.CRITICAL)
class TestForwardCurve:
    @allure.title(
        "Create a forward curve definition object with valid params and get data"
    )
    @pytest.mark.caseid("37902310")
    @pytest.mark.smoke
    def test_create_forward_curve_definition_with_valid_params(self, open_session):
        response = forward_curve_definition_01().get_data()

        check_curve_response(response, expected_currency="EUR")
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a empty forward curve definition object and get data")
    @pytest.mark.caseid("37922828")
    def test_create_empty_forward_curve_definition(self, open_session):
        with pytest.raises(
            RDError,
            match="Error code QPS-Curves.6 | Invalid input: curveDefinition is missing",
        ):
            forward_curves.Definition().get_data()

    @allure.title("Create a forward curve definition object and check curve parameters")
    @pytest.mark.caseid("37922829")
    def test_create_forward_curve_definition_and_check_curve_parameters(
        self, open_session
    ):
        response = forward_curve_definition_02().get_data()
        curve = response.data.curve
        check_curve(curve)

    @allure.title("Create a forward curve definition object async")
    @pytest.mark.caseid("37922830")
    async def test_create_forward_curve_definition_async(self, open_session_async):
        invalid_response, valid_response = await get_async_response_from_definitions(
            forward_curves.Definition(), forward_curve_definition_02()
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-Curves.6",
            expected_error_message="Invalid input: curveDefinition is missing",
        )

        curve = valid_response.data.curve
        check_curve(curve)
        assert not valid_response.data.df.empty

    @allure.title("Create a forward curve definition with extended parameters")
    @pytest.mark.caseid("C48910806")
    def test_get_forward_curve_with_extended_params(self, open_session):
        response = forward_curves.Definition(
            forward_curve_definitions=[
                forward_curves.ForwardCurveDefinition(
                    index_tenor="1M",
                    forward_curve_tenors=["0D", "1D", "11D"],
                )
            ],
            curve_definition=forward_curves.SwapZcCurveDefinition(
                currency="EUR",
                name="EUR EURIBOR Swap ZC Curve",
            ),
            curve_tag="test_curve",
            extended_params={
                "curveDefinition": {
                    "indexName": "EURIBOR",
                    "currency": "EUR",
                    "discountingTenor": "OIS",
                    "name": "EUR EURIBOR Swap ZC Curve",
                },
                "curveParameters": {"calendarAdjustment": "Calendar"},
                "forwardCurveDefinitions": [
                    {
                        "indexTenor": "3M",
                        "forwardCurveTenors": ["0D", "1D"],
                        "forwardCurveTag": "MyForwardTag",
                        "forwardStartDate": "2021-02-01",
                        "forwardStartTenor": "some_start_tenor",
                    }
                ],
                "curveTag": "my_test_curve",
            },
        ).get_data()

        check_curve_response(response, expected_currency="EUR")

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == forward_curve_universe
        ), f"Extended params are applied incorrectly"
        assert response.data.raw['data'][0]['curveTag'] == "my_test_curve"
        assert response.data.curve.forward_curve_tag == "MyForwardTag"




import pytest

import refinitiv.data as rd
from refinitiv.data._errors import RDError
from tests.unit.conftest import StubSession, StubFailedResponse


def test_search_definition():
    ############################################
    #   prepare things

    ############################################
    #   test
    definition = rd.content.search.Definition(query="foo")

    assert definition._query == "foo"

    #   check default value
    assert definition._view == rd.content.search.Views.SEARCH_ALL
    assert definition._top == 10
    assert definition._skip == 0
    assert definition._group_count == 3


def test_search_definition_with_boost_option():
    ############################################
    #   prepare things

    ############################################
    #   test
    filter = "RCSAssetTypeLeaf eq 'oil refinery' and RCSRegionLeaf eq 'Venezuela'"
    boost = "PlantStatus ne 'Normal Operation'"
    definition = rd.content.search.Definition(
        view=rd.content.search.Views.VESSEL_PHYSICAL_ASSETS,
        filter=filter,
        boost=boost,
    )

    #   check
    assert definition._view == rd.content.search.Views.VESSEL_PHYSICAL_ASSETS
    assert definition._filter == filter
    assert definition._boost == boost


def test_search_definition_with_sort_option():
    ############################################
    #   prepare things

    ############################################
    #   test
    query = "ceo"
    order_by = "YearOfBirth desc,LastName,FirstName"
    select = "YearOfBirth,DocumentTitle"
    definition = rd.content.search.Definition(
        view=rd.content.search.Views.PEOPLE,
        query=query,
        select=select,
        order_by=order_by,
    )

    #   check
    assert definition._view == rd.content.search.Views.PEOPLE
    assert definition._query == query
    assert definition._select == select
    assert definition._order_by == order_by


def test_search_definition_with_paging_option():
    ############################################
    #   prepare things

    ############################################
    #   test
    query = "gold"
    order_by = "YearOfBirth desc,LastName,FirstName"
    select = "YearOfBirth,DocumentTitle"
    definition = rd.content.search.Definition(
        view=rd.content.search.Views.COMMODITY_QUOTES,
        query=query,
        top=10,
        skip=20,
    )

    #   check
    assert definition._view == rd.content.search.Views.COMMODITY_QUOTES
    assert definition._query == query
    assert definition._top == 10
    assert definition._skip == 20


def test_raise_error_when_not_valid_request():
    # given
    session = StubSession(is_open=True, response=StubFailedResponse())
    definition = rd.content.search.Definition(view="invalid_value")

    with pytest.raises(RDError):
        definition.get_data(session=session)



from typing import List

import pandas as pd


def bulk_build_df(
    raw: dict, column_by_field: dict, columns: List[str], **kwargs
) -> pd.DataFrame:
    fields = [
        column_by_field.get(column, column) for i, (column, *_) in enumerate(columns)
    ]
    df = pd.DataFrame(raw, columns=fields)
    if not df.empty:
        df = df.convert_dtypes()

    return df



import threading
from typing import Optional

import pytest
from mock.mock import patch

from refinitiv.data import OpenState
from refinitiv.data._content_type import ContentType
from refinitiv.data.delivery._data._api_type import APIType
from refinitiv.data.delivery._stream import (
    OMMStream,
    StreamConnection,
    get_cxn_cfg_provider,
)
from refinitiv.data.delivery._stream._protocol_type import ProtocolType
from refinitiv.data.delivery._stream._stream_cxn_cache import (
    StreamCxnCache,
    stream_cxn_cache,
)
from refinitiv.data.delivery._stream._stream_factory import StreamDetails
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.delivery.stream import raw_infos
from tests.unit.delivery.stream.conftest import StubWebSocketApp
import sys

# --------------------------------------------------------------------------------------
#
# TEST CASES
#
# --------------------------------------------------------------------------------------

"""
Given opened Stream
When network failure happens
Then Cxn is disposed, the Stream is closed
"""


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_stream_closes_when_cxn_get_network_failure():
    cxn, ws = None, None
    cxn_ready = threading.Event()
    stream_cxn_cache.cxn_created.clear()

    def get_cxn_ready():
        nonlocal cxn, ws

        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws

        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")

    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    assert omm_stream.open_state is OpenState.Opened

    network_failure_event = threading.Event()

    def on_network_failure():
        network_failure_event.set()
        ws.cmd("on_close")

    threading.Thread(
        name=f"{func_name}-3", target=on_network_failure, daemon=True
    ).start()

    network_failure_event.wait()
    cxn.start_disconnecting()
    cxn.end_disconnecting()
    cxn.dispose()
    cxn.join()

    assert omm_stream.open_state is OpenState.Closed

    assert stream_cxn_cache.get_cxns(session) == []

    session.close()


# --------------------------------------------------------------------------------------

"""
Given the Stream is connecting
When user close the Session
Then Cxn is disposed, the Stream is opened
And user can close the Stream
"""


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_close_session_when_stream_is_connecting():
    def close_session():
        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        session.close()

    stream_cxn_cache.cxn_created.clear()

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=close_session, daemon=True).start()
    omm_stream.open()

    assert omm_stream.open_state is OpenState.Opened

    assert stream_cxn_cache.get_cxns(session) == []

    omm_stream.close()

    assert omm_stream.open_state is OpenState.Closed

    session.close()


"""
Given the Stream is opened
When user close the Session
Then Cxn is disposed, the Stream is closed
"""


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_granted_then_close_session():
    ws, cxn = None, None
    cxn_ready = threading.Event()

    def get_cxn_ready():
        nonlocal cxn, ws

        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws

        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    stream_cxn_cache.cxn_created.clear()

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    assert omm_stream.open_state is OpenState.Opened

    assert stream_cxn_cache.get_cxns(session) != []

    session.close()

    assert stream_cxn_cache.get_cxns(session) == []

    assert omm_stream.open_state is OpenState.Closed


# --------------------------------------------------------------------------------------

"""
Given the Stream is connecting
When server response "State - Stream is Open"
Then Cxn is in message processing state
"""


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_granted_when_stream_is_opening():
    cxn_ready = threading.Event()
    cxn, ws = None, None

    def get_cxn_ready():
        nonlocal cxn, ws

        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    stream_cxn_cache.cxn_created.clear()

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    assert stream_cxn_cache.get_cxns(session) != []

    assert omm_stream.open_state is OpenState.Opened

    session.close()


# Cache perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_granted_when_cache_get_cxn():
    def get_response():
        cache.cxn_created.wait()
        cxn, *_ = cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)

    session = StubSession(response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    details = StreamDetails(
        ContentType.STREAMING_PRICING,
        ProtocolType.OMM,
        APIType.STREAMING_PRICING,
    )
    cache = StreamCxnCache()
    cache.cxn_created.clear()

    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_response, daemon=True).start()
    cxn = cache.get_cxn(session, details)

    assert cxn.is_message_processing, cxn.state

    # teardown
    cxn.start_disconnecting()
    cxn.end_disconnecting()
    cxn.dispose()
    cxn.join()
    session.close()


# --------------------------------------------------------------------------------------

"""
Given the Stream is connecting
When server response "State - Stream is Closed"
Or server response "State - Text is Login Rejected."
Or server response "State - Code is UserUnknownToPermSys"
Then Cxn is disposed
And raise ConnectionError
"""


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@pytest.mark.parametrize(
    "message",
    [
        '[{"State": {"Stream": "Closed"}}]',
        '[{"State": {"Text": "Login Rejected."}}]',
        '[{"State": {"Text": "Login Rejected."}}]',
    ],
)
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_denied_when_stream_is_opening_will_raise_error(message):
    cxn: Optional["StreamConnection"] = None
    stream_cxn_cache.cxn_created.clear()

    def get_cxn_result():
        nonlocal cxn
        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        ws.cmd("on_message", message)

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_cxn_result, daemon=True).start()

    try:
        omm_stream.open()
    except ConnectionError:
        assert True
    else:
        assert False

    assert stream_cxn_cache.get_cxns(session) == []

    assert cxn.is_disposed is True, cxn.state

    assert omm_stream.open_state is OpenState.Pending

    # teardown
    cxn.start_disconnecting()
    cxn.end_disconnecting()
    cxn.dispose()
    cxn.join()
    session.close()


# Cache perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_denied_when_cache_get_cxn_will_raise_error():
    cxn = None

    def stream_closed_message():
        nonlocal cxn
        cache.cxn_created.wait()
        cxn, *_ = cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Closed"}}]'
        ws.cmd("on_message", message)

    from tests.unit.delivery.stream.raw_infos import data

    session = StubSession(response=StubResponse(data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    details = StreamDetails(
        ContentType.STREAMING_PRICING,
        ProtocolType.OMM,
        APIType.STREAMING_PRICING,
    )
    cache = StreamCxnCache()
    cache.cxn_created.clear()

    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=stream_closed_message, daemon=True).start()
    try:
        cache.get_cxn(session, details)
    except ConnectionError:
        assert cxn.is_disposed, cxn.state
    else:
        assert False

    # teardown
    cxn.start_disconnecting()
    cxn.end_disconnecting()
    cxn.dispose()
    cxn.join()
    session.close()


# --------------------------------------------------------------------------------------

# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_granted_then_close_stream_then_close_session():
    ws, cxn = None, None
    cxn_ready = threading.Event()

    def get_cxn_ready():
        nonlocal cxn, ws

        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    stream_cxn_cache.cxn_created.clear()

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    omm_stream.close()

    session.close()

    assert stream_cxn_cache.get_cxns(session) == []

    assert omm_stream.open_state is OpenState.Closed

    # teardown
    cxn.start_disconnecting()
    cxn.end_disconnecting()
    cxn.dispose()
    cxn.join()


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_granted_then_close_stream_then_close_session():
    ws, cxn = None, None
    cxn_ready = threading.Event()

    def get_cxn_ready():
        nonlocal cxn, ws

        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    stream_cxn_cache.cxn_created.clear()

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    session.close()

    omm_stream.close()

    assert stream_cxn_cache.get_cxns(session) == []

    assert omm_stream.open_state is OpenState.Closed

    # teardown
    cxn.start_disconnecting()
    cxn.end_disconnecting()
    cxn.dispose()
    cxn.join()


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_access_granted_then_close_stream_cxn_is_in_message_processing():
    ws, cxn = None, None
    cxn_ready = threading.Event()

    def get_cxn_ready():
        nonlocal cxn, ws

        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    stream_cxn_cache.cxn_created.clear()

    session = StubSession(is_open=True, response=StubResponse(raw_infos.data))
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    omm_stream.close()

    cxn, *_ = stream_cxn_cache.get_cxns(session)

    assert cxn.is_message_processing is True, cxn.state

    # teardown
    cxn.start_disconnecting()
    cxn.end_disconnecting()
    cxn.dispose()
    cxn.join()
    session.close()
